# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

# include parent path
import os
import sys
import numpy as np
import math
import ctypes

import unittest

import warp as wp
from warp.tests.test_base import *

#wp.config.mode = "debug"

wp.init()

@wp.kernel
def test_rename():


    a = 0
    b = 1
    
    a = b
    a = 2

    wp.expect_eq(a, 2)
    wp.expect_eq(b, 1)
    

@wp.kernel
def test_inplace():

    a = 1.0
    a += 2.0
    
    wp.expect_eq(a, 3.0)


@wp.kernel
def test_constant(c: float):

    a = 0.0
    a = c + 1.0

    wp.expect_eq(a, 2.0)

@wp.kernel
def test_dynamic_for_rename(n: int):

    f0 = int(0.0)
    f1 = int(1.0)

    for i in range(0, n):
        
        f = f0 + f1
        
        f0 = f1
        f1 = f

    wp.expect_eq(f1, 89)

@wp.kernel
def test_dynamic_for_inplace(n: int):

    a = float(0.0)

    for i in range(0, n):
        a += 1.0

    wp.expect_eq(n, 10)


@wp.kernel
def test_reassign():

    f0 = 1.0
    f1 = f0

    f1 = f1 + 2.0

    wp.expect_eq(f1, 3.0)
    wp.expect_eq(f0, 1.0)

@wp.kernel
def test_dynamic_reassign(n: int):

    f0 = wp.vec3()
    f1 = f0

    for i in range(0, n):
        f1 = f1 - wp.vec3(2.0, 0.0, 0.0)

    wp.expect_eq(f1, wp.vec3(-4.0, 0.0, 0.0))
    wp.expect_eq(f0, wp.vec3())


@wp.kernel
def test_range_static_sum(result: wp.array(dtype=int)):

    a = int(0)
    for i in range(10):
        a = a + 1

    b = int(0)
    for i in range(0, 10):
        b = b + 1

    c = int(0)
    for i in range(0, 20, 2):
        c = c + 1

    result[0] = a
    result[1] = b
    result[2] = c


@wp.kernel
def test_range_dynamic_sum(start: int, end: int, step: int, result: wp.array(dtype=int)):

    a = int(0)
    for i in range(end):
        a = a + 1

    b = int(0)
    for i in range(start, end):
        b = b + 1

    c = int(0)
    for i in range(start, end*step, step):
        c = c + 1

    d = int(0)
    for i in range(end*step, start, -step):
        d = d + 1

    result[0] = a
    result[1] = b
    result[2] = c
    result[3] = d

@wp.kernel
def test_range_dynamic(start: int, end: int, step: int, result: wp.array(dtype=int)):

    output = int(0)
    for i in range(start, end, step):
        result[output] = i
        output += 1

@wp.kernel
def test_range_dynamic_nested(n: int):
    
    sum1 = float(0.0)
    sum2 = float(0.0)
    sum3 = float(0.0)

    for i in range(n):
        sum1 = sum1 + 1.0
        sum3 = sum3 + 1.0

        for j in range(n):
            sum2 = sum2 + 1.0
            sum3 = sum3 + 1.0
        
        sum3 = sum3 + 1.0


    wp.expect_eq(sum1, float(n))
    wp.expect_eq(sum2, float(n*n))
    wp.expect_eq(sum3, float(n*n + 2*n))



@wp.kernel
def test_while(n: int):

    i = int(0)

    while i < n:
        i = i + 1

    wp.expect_eq(i, n)


def register(parent):

    class TestCodeGen(parent):
        pass

    devices = wp.get_devices()

    add_kernel_test(TestCodeGen, name="test_inplace", kernel=test_inplace, dim=1, devices=devices)
    add_kernel_test(TestCodeGen, name="test_rename", kernel=test_rename, dim=1, devices=devices)
    add_kernel_test(TestCodeGen, name="test_constant", kernel=test_constant, inputs=[1.0], dim=1, devices=devices)
    add_kernel_test(TestCodeGen, name="test_dynamic_for_rename", kernel=test_dynamic_for_rename, inputs=[10], dim=1, devices=devices)
    add_kernel_test(TestCodeGen, name="test_dynamic_for_inplace", kernel=test_dynamic_for_inplace, inputs=[10], dim=1, devices=devices)
    add_kernel_test(TestCodeGen, name="test_reassign", kernel=test_reassign, dim=1, devices=devices)
    add_kernel_test(TestCodeGen, name="test_dynamic_reassign", kernel=test_dynamic_reassign, inputs=[2], dim=1, devices=devices)

    add_kernel_test(TestCodeGen, name="test_range_dynamic_forward", kernel=test_range_dynamic, dim=1, inputs=[0, 4, 1], expect=[0, 1, 2, 3], devices=devices)
    add_kernel_test(TestCodeGen, name="test_range_dynamic_reverse", kernel=test_range_dynamic, dim=1, inputs=[4, 0, -1], expect=[4, 3, 2, 1], devices=devices)
    add_kernel_test(TestCodeGen, name="test_range_dynamic_foward_step", kernel=test_range_dynamic, dim=1, inputs=[0, 8, 2], expect=[0, 2, 4, 6], devices=devices)
    add_kernel_test(TestCodeGen, name="test_range_dynamic_reverse_step", kernel=test_range_dynamic, dim=1, inputs=[8, 0, -2], expect=[8, 6, 4, 2], devices=devices)

    add_kernel_test(TestCodeGen, name="test_range_static_sum", kernel=test_range_static_sum, dim=1, expect=[10, 10, 10], devices=devices)
    add_kernel_test(TestCodeGen, name="test_range_dynamic_sum", kernel=test_range_dynamic_sum, dim=1, inputs=[0, 10, 2], expect=[10, 10, 10, 10], devices=devices)
    add_kernel_test(TestCodeGen, name="test_range_dynamic_sum_zero", kernel=test_range_dynamic_sum, dim=1, inputs=[0, 0, 1], expect=[0, 0, 0, 0], devices=devices)

    add_kernel_test(TestCodeGen, name="test_range_dynamic_nested", kernel=test_range_dynamic_nested, dim=1, inputs=[4], devices=devices)

    add_kernel_test(TestCodeGen, name="test_while_zero", kernel=test_while, dim=1, inputs=[0], devices=devices)
    add_kernel_test(TestCodeGen, name="test_while_positive", kernel=test_while, dim=1, inputs=[16], devices=devices)

    return TestCodeGen


if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2)





# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

# include parent path
import numpy as np
import math

import warp as wp
from warp.tests.test_base import *

np.random.seed(42)

wp.init()
wp.verify_cuda = True

@wp.kernel
def test_print():

    wp.print(1.0)
    wp.print("this is a string")
    wp.printf("this is a float %f\n", 457.5)
    wp.printf("this is an int %d\n", 123)


wp.launch(
    kernel=test_print,
    dim=1,
    inputs=[],
    outputs=[],
    device="cuda")

wp.synchronize()
print("finished")




import warp as wp

# dummy class used in test_reload.py
class ClassKernelTest:

    def __init__(self):
        
        # 3x3 frames in the rest pose:
        self.identities = wp.zeros(shape=10,dtype=wp.mat33, device="cuda:0")
        wp.launch(
            kernel=self.gen_identities_kernel, dim=10, inputs=[self.identities],
        )

    @wp.func
    def return_identity( e: int ):
        return wp.mat33(1.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,1.0)

    @wp.kernel
    def gen_identities_kernel(s: wp.array(dtype=wp.mat33)):
        tid = wp.tid()
        s[tid] = ClassKernelTest.return_identity(tid)



# Autogenerated file, do not edit, this file provides stubs for builtins autocomplete in VSCode, PyCharm, etc

from typing import Any
from typing import Tuple
from typing import Callable
from typing import overload
from warp.types import array, array2d, array3d, array4d, constant
from warp.types import int8, uint8, int16, uint16, int32, uint32, int64, uint64, float16, float32, float64
from warp.types import vec2, vec3, vec4, mat22, mat33, mat44, quat, transform, spatial_vector, spatial_matrix
from warp.types import bvh_query_t, mesh_query_aabb_t, hash_grid_query_t


@overload
def min(x: int32, y: int32) -> int:
   """
   Return the minimum of two integers.
   """
   ...

@overload
def min(x: float32, y: float32) -> float:
   """
   Return the minimum of two floats.
   """
   ...

@overload
def max(x: int32, y: int32) -> int:
   """
   Return the maximum of two integers.
   """
   ...

@overload
def max(x: float32, y: float32) -> float:
   """
   Return the maximum of two floats.
   """
   ...

@overload
def clamp(x: int32, a: int32, b: int32) -> int:
   """
   Clamp the value of x to the range [a, b].
   """
   ...

@overload
def clamp(x: float32, a: float32, b: float32) -> float:
   """
   Clamp the value of x to the range [a, b].
   """
   ...

@overload
def abs(x: int32) -> int:
   """
   Return the absolute value of x.
   """
   ...

@overload
def abs(x: float32) -> float:
   """
   Return the absolute value of x.
   """
   ...

@overload
def sign(x: int32) -> int:
   """
   Return -1 if x < 0, return 1 otherwise.
   """
   ...

@overload
def sign(x: float32) -> float:
   """
   Return -1.0 if x < 0.0, return 1.0 otherwise.
   """
   ...

@overload
def step(x: float32) -> float:
   """
   Return 1.0 if x < 0.0, return 0.0 otherwise.
   """
   ...

@overload
def nonzero(x: float32) -> float:
   """
   Return 1.0 if x is not equal to zero, return 0.0 otherwise.
   """
   ...

@overload
def sin(x: float32) -> float:
   """
   Return the sine of x in radians.
   """
   ...

@overload
def cos(x: float32) -> float:
   """
   Return the cosine of x in radians.
   """
   ...

@overload
def acos(x: float32) -> float:
   """
   Return arccos of x in radians. Inputs are automatically clamped to [-1.0, 1.0].
   """
   ...

@overload
def asin(x: float32) -> float:
   """
   Return arcsin of x in radians. Inputs are automatically clamped to [-1.0, 1.0].
   """
   ...

@overload
def sqrt(x: float32) -> float:
   """
   Return the sqrt of x, where x is positive.
   """
   ...

@overload
def tan(x: float32) -> float:
   """
   Return tangent of x in radians.
   """
   ...

@overload
def atan(x: float32) -> float:
   """
   Return arctan of x.
   """
   ...

@overload
def atan2(y: float32, x: float32) -> float:
   """
   Return atan2 of x.
   """
   ...

@overload
def sinh(x: float32) -> float:
   """
   Return the sinh of x.
   """
   ...

@overload
def cosh(x: float32) -> float:
   """
   Return the cosh of x.
   """
   ...

@overload
def tanh(x: float32) -> float:
   """
   Return the tanh of x.
   """
   ...

@overload
def log(x: float32) -> float:
   """
   Return the natural log (base-e) of x, where x is positive.
   """
   ...

@overload
def log2(x: float32) -> float:
   """
   Return the natural log (base-2) of x, where x is positive.
   """
   ...

@overload
def log10(x: float32) -> float:
   """
   Return the natural log (base-10) of x, where x is positive.
   """
   ...

@overload
def exp(x: float32) -> float:
   """
   Return base-e exponential, e^x.
   """
   ...

@overload
def pow(x: float32, y: float32) -> float:
   """
   Return the result of x raised to power of y.
   """
   ...

@overload
def round(x: float32) -> float:
   """
   Calculate the nearest integer value, rounding halfway cases away from zero.
      This is the most intuitive form of rounding in the colloquial sense, but can be slower than other options like ``warp.rint()``.
      Differs from ``numpy.round()``, which behaves the same way as ``numpy.rint()``.
   """
   ...

@overload
def rint(x: float32) -> float:
   """
   Calculate the nearest integer value, rounding halfway cases to nearest even integer.
      It is generally faster than ``warp.round()``.
      Equivalent to ``numpy.rint()``.
   """
   ...

@overload
def trunc(x: float32) -> float:
   """
   Calculate the nearest integer that is closer to zero than x.
      In other words, it discards the fractional part of x.
      It is similar to casting ``float(int(x))``, but preserves the negative sign when x is in the range [-0.0, -1.0).
      Equivalent to ``numpy.trunc()`` and ``numpy.fix()``.
   """
   ...

@overload
def floor(x: float32) -> float:
   """
   Calculate the largest integer that is less than or equal to x.
   """
   ...

@overload
def ceil(x: float32) -> float:
   """
   Calculate the smallest integer that is greater than or equal to x.
   """
   ...

@overload
def dot(x: vec2, y: vec2) -> float:
   """
   Compute the dot product between two 2d vectors.
   """
   ...

@overload
def dot(x: vec3, y: vec3) -> float:
   """
   Compute the dot product between two 3d vectors.
   """
   ...

@overload
def dot(x: vec4, y: vec4) -> float:
   """
   Compute the dot product between two 4d vectors.
   """
   ...

@overload
def dot(x: quat, y: quat) -> float:
   """
   Compute the dot product between two quaternions.
   """
   ...

@overload
def outer(x: vec2, y: vec2) -> mat22:
   """
   Compute the outer product x*y^T for two vec2 objects.
   """
   ...

@overload
def outer(x: vec3, y: vec3) -> mat33:
   """
   Compute the outer product x*y^T for two vec3 objects.
   """
   ...

@overload
def cross(x: vec3, y: vec3) -> vec3:
   """
   Compute the cross product of two 3d vectors.
   """
   ...

@overload
def skew(x: vec3) -> mat33:
   """
   Compute the skew symmetric matrix for a 3d vector.
   """
   ...

@overload
def length(x: vec2) -> float:
   """
   Compute the length of a 2d vector.
   """
   ...

@overload
def length(x: vec3) -> float:
   """
   Compute the length of a 3d vector.
   """
   ...

@overload
def length(x: vec4) -> float:
   """
   Compute the length of a 4d vector.
   """
   ...

@overload
def length(x: quat) -> float:
   """
   Compute the length of a quaternion.
   """
   ...

@overload
def length_sq(x: vec2) -> float:
   """
   Compute the squared length of a 2d vector.
   """
   ...

@overload
def length_sq(x: vec3) -> float:
   """
   Compute the squared length of a 3d vector.
   """
   ...

@overload
def length_sq(x: vec4) -> float:
   """
   Compute the squared length of a 4d vector.
   """
   ...

@overload
def length_sq(x: quat) -> float:
   """
   Compute the squared length of a quaternion.
   """
   ...

@overload
def normalize(x: vec2) -> vec2:
   """
   Compute the normalized value of x, if length(x) is 0 then the zero vector is returned.
   """
   ...

@overload
def normalize(x: vec3) -> vec3:
   """
   Compute the normalized value of x, if length(x) is 0 then the zero vector is returned.
   """
   ...

@overload
def normalize(x: vec4) -> vec4:
   """
   Compute the normalized value of x, if length(x) is 0 then the zero vector is returned.
   """
   ...

@overload
def normalize(x: quat) -> quat:
   """
   Compute the normalized value of x, if length(x) is 0 then the zero quat is returned.
   """
   ...

@overload
def transpose(m: mat22) -> mat22:
   """
   Return the transpose of the matrix m
   """
   ...

@overload
def transpose(m: mat33) -> mat33:
   """
   Return the transpose of the matrix m
   """
   ...

@overload
def transpose(m: mat44) -> mat44:
   """
   Return the transpose of the matrix m
   """
   ...

@overload
def transpose(m: spatial_matrix) -> spatial_matrix:
   """
   Return the transpose of the matrix m
   """
   ...

@overload
def inverse(m: mat22) -> mat22:
   """
   Return the inverse of the matrix m
   """
   ...

@overload
def inverse(m: mat33) -> mat33:
   """
   Return the inverse of the matrix m
   """
   ...

@overload
def inverse(m: mat44) -> mat44:
   """
   Return the inverse of the matrix m
   """
   ...

@overload
def determinant(m: mat22) -> float:
   """
   Return the determinant of the matrix m
   """
   ...

@overload
def determinant(m: mat33) -> float:
   """
   Return the determinant of the matrix m
   """
   ...

@overload
def determinant(m: mat44) -> float:
   """
   Return the determinant of the matrix m
   """
   ...

@overload
def trace(m: mat22) -> float:
   """
   Return the trace of the matrix m
   """
   ...

@overload
def trace(m: mat33) -> float:
   """
   Return the trace of the matrix m
   """
   ...

@overload
def trace(m: mat44) -> float:
   """
   Return the trace of the matrix m
   """
   ...

@overload
def diag(d: vec2) -> mat22:
   """
   Returns a matrix with the components of the vector d on the diagonal
   """
   ...

@overload
def diag(d: vec3) -> mat33:
   """
   Returns a matrix with the components of the vector d on the diagonal
   """
   ...

@overload
def diag(d: vec4) -> mat44:
   """
   Returns a matrix with the components of the vector d on the diagonal
   """
   ...

@overload
def cw_mul(x: vec2, y: vec2) -> vec2:
   """
   Component wise multiply of two 2d vectors.
   """
   ...

@overload
def cw_mul(x: vec3, y: vec3) -> vec3:
   """
   Component wise multiply of two 3d vectors.
   """
   ...

@overload
def cw_mul(x: vec4, y: vec4) -> vec4:
   """
   Component wise multiply of two 4d vectors.
   """
   ...

@overload
def cw_div(x: vec2, y: vec2) -> vec2:
   """
   Component wise division of two 2d vectors.
   """
   ...

@overload
def cw_div(x: vec3, y: vec3) -> vec3:
   """
   Component wise division of two 3d vectors.
   """
   ...

@overload
def cw_div(x: vec4, y: vec4) -> vec4:
   """
   Component wise division of two 4d vectors.
   """
   ...

@overload
def svd3(A: mat33, U: mat33, sigma: vec3, V: mat33):
   """
   Compute the SVD of a 3x3 matrix. The singular values are returned in sigma, 
      while the left and right basis vectors are returned in U and V.
   """
   ...

@overload
def quat_identity() -> quat:
   """
   Construct an identity quaternion with zero imaginary part and real part of 1.0
   """
   ...

@overload
def quat_from_axis_angle(axis: vec3, angle: float32) -> quat:
   """
   Construct a quaternion representing a rotation of angle radians around the given axis.
   """
   ...

@overload
def quat_from_matrix(m: mat33) -> quat:
   """
   Construct a quaternion from a 3x3 matrix.
   """
   ...

@overload
def quat_rpy(roll: float32, pitch: float32, yaw: float32) -> quat:
   """
   Construct a quaternion representing a combined roll (z), pitch (x), yaw rotations (y) in radians.
   """
   ...

@overload
def quat_inverse(q: quat) -> quat:
   """
   Compute quaternion conjugate.
   """
   ...

@overload
def quat_rotate(q: quat, p: vec3) -> vec3:
   """
   Rotate a vector by a quaternion.
   """
   ...

@overload
def quat_rotate_inv(q: quat, p: vec3) -> vec3:
   """
   Rotate a vector the inverse of a quaternion.
   """
   ...

@overload
def quat_to_matrix(q: quat) -> mat33:
   """
   Convert a quaternion to a 3x3 rotation matrix.
   """
   ...

@overload
def transform_identity() -> transform:
   """
   Construct an identity transform with zero translation and identity rotation.
   """
   ...

@overload
def transform_get_translation(t: transform) -> vec3:
   """
   Return the translational part of a transform.
   """
   ...

@overload
def transform_get_rotation(t: transform) -> quat:
   """
   Return the rotational part of a transform.
   """
   ...

@overload
def transform_multiply(a: transform, b: transform) -> transform:
   """
   Multiply two rigid body transformations together.
   """
   ...

@overload
def transform_point(t: transform, p: vec3) -> vec3:
   """
   Apply the transform to a point p treating the homogenous coordinate as w=1 (translation and rotation).
   """
   ...

@overload
def transform_point(m: mat44, p: vec3) -> vec3:
   """
   Apply the transform to a point ``p`` treating the homogenous coordinate as w=1. The transformation is applied treating ``p`` as a column vector, e.g.: ``y = M*p``
      note this is in contrast to some libraries, notably USD, which applies transforms to row vectors, ``y^T = p^T*M^T``. If the transform is coming from a library that uses row-vectors
      then users should transpose the tranformation matrix before calling this method.
   """
   ...

@overload
def transform_vector(t: transform, v: vec3) -> vec3:
   """
   Apply the transform to a vector v treating the homogenous coordinate as w=0 (rotation only).
   """
   ...

@overload
def transform_vector(m: mat44, v: vec3) -> vec3:
   """
   Apply the transform to a vector ``v`` treating the homogenous coordinate as w=0. The transformation is applied treating ``v`` as a column vector, e.g.: ``y = M*v``
      note this is in contrast to some libraries, notably USD, which applies transforms to row vectors, ``y^T = v^T*M^T``. If the transform is coming from a library that uses row-vectors
      then users should transpose the tranformation matrix before calling this method.
   """
   ...

@overload
def transform_inverse(t: transform) -> transform:
   """
   Compute the inverse of the transform.
   """
   ...

@overload
def spatial_dot(a: spatial_vector, b: spatial_vector) -> float:
   """
   Compute the dot product of two 6d screw vectors.
   """
   ...

@overload
def spatial_cross(a: spatial_vector, b: spatial_vector) -> spatial_vector:
   """
   Compute the cross-product of two 6d screw vectors.
   """
   ...

@overload
def spatial_cross_dual(a: spatial_vector, b: spatial_vector) -> spatial_vector:
   """
   Compute the dual cross-product of two 6d screw vectors.
   """
   ...

@overload
def spatial_top(a: spatial_vector) -> vec3:
   """
   Return the top (first) part of a 6d screw vector.
   """
   ...

@overload
def spatial_bottom(a: spatial_vector) -> vec3:
   """
   Return the bottom (second) part of a 6d screw vector.
   """
   ...

@overload
def spatial_jacobian(S: array[spatial_vector], joint_parents: array[int32], joint_qd_start: array[int32], joint_start: int32, joint_count: int32, J_start: int32, J_out: array[float32]):
   """

   """
   ...

@overload
def spatial_mass(I_s: array[spatial_matrix], joint_start: int32, joint_count: int32, M_start: int32, M: array[float32]):
   """

   """
   ...

@overload
def mlp(weights: array[float32], bias: array[float32], activation: Callable, index: int32, x: array[float32], out: array[float32]):
   """
   Evaluate a multi-layer perceptron (MLP) layer in the form: ``out = act(weights*x + bias)``. 

      :param weights: A layer's network weights with dimensions ``(m, n)``.
      :param bias: An array with dimensions ``(n)``.
      :param activation: A ``wp.func`` function that takes a single scalar float as input and returns a scalar float as output
      :param index: The batch item to process, typically each thread will process 1 item in the batch, in this case index should be ``wp.tid()``
      :param x: The feature matrix with dimensions ``(n, b)``
      :param out: The network output with dimensions ``(m, b)``

      :note: Feature and output matrices are transposed compared to some other frameworks such as PyTorch. All matrices are assumed to be stored in flattened row-major memory layout (NumPy default).
   """
   ...

@overload
def bvh_query_aabb(id: uint64, lower: vec3, upper: vec3) -> bvh_query_t:
   """
   Construct an axis-aligned bounding box query against a bvh object. This query can be used to iterate over all bounds
      inside a bvh. Returns an object that is used to track state during bvh traversal.
    
      :param id: The bvh identifier
      :param lower: The lower bound of the bounding box in bvh space
      :param upper: The upper bound of the bounding box in bvh space
   """
   ...

@overload
def bvh_query_ray(id: uint64, start: vec3, dir: vec3) -> bvh_query_t:
   """
   Construct a ray query against a bvh object. This query can be used to iterate over all bounds
      that intersect the ray. Returns an object that is used to track state during bvh traversal.
    
      :param id: The bvh identifier
      :param start: The start of the ray in bvh space
      :param dir: The direction of the ray in bvh space
   """
   ...

@overload
def bvh_query_next(query: bvh_query_t, index: int32) -> bool:
   """
   Move to the next bound returned by the query. The index of the current bound is stored in ``index``, returns ``False``
      if there are no more overlapping bound.
   """
   ...

@overload
def mesh_query_point(id: uint64, point: vec3, max_dist: float32, inside: float32, face: int32, bary_u: float32, bary_v: float32) -> bool:
   """
   Computes the closest point on the mesh with identifier `id` to the given point in space. Returns ``True`` if a point < ``max_dist`` is found.

      :param id: The mesh identifier
      :param point: The point in space to query
      :param max_dist: Mesh faces above this distance will not be considered by the query
      :param inside: Returns a value < 0 if query point is inside the mesh, >=0 otherwise. Note that mesh must be watertight for this to be robust
      :param face: Returns the index of the closest face
      :param bary_u: Returns the barycentric u coordinate of the closest point
      :param bary_v: Retruns the barycentric v coordinate of the closest point
   """
   ...

@overload
def mesh_query_ray(id: uint64, start: vec3, dir: vec3, max_t: float32, t: float32, bary_u: float32, bary_v: float32, sign: float32, normal: vec3, face: int32) -> bool:
   """
   Computes the closest ray hit on the mesh with identifier `id`, returns ``True`` if a point < ``max_t`` is found.

      :param id: The mesh identifier
      :param start: The start point of the ray
      :param dir: The ray direction (should be normalized)
      :param max_t: The maximum distance along the ray to check for intersections
      :param t: Returns the distance of the closest hit along the ray
      :param bary_u: Returns the barycentric u coordinate of the closest hit
      :param bary_v: Returns the barycentric v coordinate of the closest hit
      :param sign: Returns a value > 0 if the hit ray hit front of the face, returns < 0 otherwise
      :param normal: Returns the face normal
      :param face: Returns the index of the hit face
   """
   ...

@overload
def mesh_query_aabb(id: uint64, lower: vec3, upper: vec3) -> mesh_query_aabb_t:
   """
   Construct an axis-aligned bounding box query against a mesh object. This query can be used to iterate over all triangles
      inside a volume. Returns an object that is used to track state during mesh traversal.
    
      :param id: The mesh identifier
      :param lower: The lower bound of the bounding box in mesh space
      :param upper: The upper bound of the bounding box in mesh space
   """
   ...

@overload
def mesh_query_aabb_next(query: mesh_query_aabb_t, index: int32) -> bool:
   """
   Move to the next triangle overlapping the query bounding box. The index of the current face is stored in ``index``, returns ``False``
      if there are no more overlapping triangles.
   """
   ...

@overload
def mesh_eval_position(id: uint64, face: int32, bary_u: float32, bary_v: float32) -> vec3:
   """
   Evaluates the position on the mesh given a face index, and barycentric coordinates.
   """
   ...

@overload
def mesh_eval_velocity(id: uint64, face: int32, bary_u: float32, bary_v: float32) -> vec3:
   """
   Evaluates the velocity on the mesh given a face index, and barycentric coordinates.
   """
   ...

@overload
def hash_grid_query(id: uint64, point: vec3, max_dist: float32) -> hash_grid_query_t:
   """
   Construct a point query against a hash grid. This query can be used to iterate over all neighboring points withing a 
      fixed radius from the query point. Returns an object that is used to track state during neighbor traversal.
   """
   ...

@overload
def hash_grid_query_next(query: hash_grid_query_t, index: int32) -> bool:
   """
   Move to the next point in the hash grid query. The index of the current neighbor is stored in ``index``, returns ``False``
      if there are no more neighbors.
   """
   ...

@overload
def hash_grid_point_id(id: uint64, index: int32) -> int:
   """
   Return the index of a point in the grid, this can be used to re-order threads such that grid 
      traversal occurs in a spatially coherent order.
   """
   ...

@overload
def intersect_tri_tri(v0: vec3, v1: vec3, v2: vec3, u0: vec3, u1: vec3, u2: vec3) -> int:
   """
   Tests for intersection between two triangles (v0, v1, v2) and (u0, u1, u2) using Moller's method. Returns > 0 if triangles intersect.
   """
   ...

@overload
def mesh_eval_face_normal(id: uint64, face: int32) -> vec3:
   """
   Evaluates the face normal the mesh given a face index.
   """
   ...

@overload
def mesh_get_point(id: uint64, index: int32) -> vec3:
   """
   Returns the point of the mesh given a index.
   """
   ...

@overload
def mesh_get_velocity(id: uint64, index: int32) -> vec3:
   """
   Returns the velocity of the mesh given a index.
   """
   ...

@overload
def mesh_get_index(id: uint64, index: int32) -> int:
   """
   Returns the point-index of the mesh given a face-vertex index.
   """
   ...

@overload
def closest_point_edge_edge(p1: vec3, q1: vec3, p2: vec3, q2: vec3, epsilon: float32) -> vec3:
   """
   Finds the closest points between two edges. Returns barycentric weights to the points on each edge, as well as the closest distance between the edges.

      :param p1: First point of first edge
      :param q1: Second point of first edge
      :param p2: First point of second edge
      :param q2: Second point of second edge
      :param epsilon: Zero tolerance for determining if points in an edge are degenerate.
      :param out: vec3 output containing (s,t,d), where `s` in [0,1] is the barycentric weight for the first edge, `t` is the barycentric weight for the second edge, and `d` is the distance between the two edges at these two closest points.
   """
   ...

@overload
def volume_sample_f(id: uint64, uvw: vec3, sampling_mode: int32) -> float:
   """
   Sample the volume given by ``id`` at the volume local-space point ``uvw``. Interpolation should be ``wp.Volume.CLOSEST``, or ``wp.Volume.LINEAR.``
   """
   ...

@overload
def volume_lookup_f(id: uint64, i: int32, j: int32, k: int32) -> float:
   """
   Returns the value of voxel with coordinates ``i``, ``j``, ``k``, if the voxel at this index does not exist this function returns the background value
   """
   ...

@overload
def volume_store_f(id: uint64, i: int32, j: int32, k: int32, value: float32):
   """
   Store the value at voxel with coordinates ``i``, ``j``, ``k``.
   """
   ...

@overload
def volume_sample_v(id: uint64, uvw: vec3, sampling_mode: int32) -> vec3:
   """
   Sample the vector volume given by ``id`` at the volume local-space point ``uvw``. Interpolation should be ``wp.Volume.CLOSEST``, or ``wp.Volume.LINEAR.``
   """
   ...

@overload
def volume_lookup_v(id: uint64, i: int32, j: int32, k: int32) -> vec3:
   """
   Returns the vector value of voxel with coordinates ``i``, ``j``, ``k``, if the voxel at this index does not exist this function returns the background value
   """
   ...

@overload
def volume_store_v(id: uint64, i: int32, j: int32, k: int32, value: vec3):
   """
   Store the value at voxel with coordinates ``i``, ``j``, ``k``.
   """
   ...

@overload
def volume_sample_i(id: uint64, uvw: vec3) -> int:
   """
   Sample the int32 volume given by ``id`` at the volume local-space point ``uvw``. 
   """
   ...

@overload
def volume_lookup_i(id: uint64, i: int32, j: int32, k: int32) -> int:
   """
   Returns the int32 value of voxel with coordinates ``i``, ``j``, ``k``, if the voxel at this index does not exist this function returns the background value
   """
   ...

@overload
def volume_index_to_world(id: uint64, uvw: vec3) -> vec3:
   """
   Transform a point defined in volume index space to world space given the volume's intrinsic affine transformation.
   """
   ...

@overload
def volume_world_to_index(id: uint64, xyz: vec3) -> vec3:
   """
   Transform a point defined in volume world space to the volume's index space, given the volume's intrinsic affine transformation.
   """
   ...

@overload
def volume_index_to_world_dir(id: uint64, uvw: vec3) -> vec3:
   """
   Transform a direction defined in volume index space to world space given the volume's intrinsic affine transformation.
   """
   ...

@overload
def volume_world_to_index_dir(id: uint64, xyz: vec3) -> vec3:
   """
   Transform a direction defined in volume world space to the volume's index space, given the volume's intrinsic affine transformation.
   """
   ...

@overload
def rand_init(seed: int32) -> uint32:
   """
   Initialize a new random number generator given a user-defined seed. Returns a 32-bit integer representing the RNG state.
   """
   ...

@overload
def rand_init(seed: int32, offset: int32) -> uint32:
   """
   Initialize a new random number generator given a user-defined seed and an offset. 
      This alternative constructor can be useful in parallel programs, where a kernel as a whole should share a seed,
      but each thread should generate uncorrelated values. In this case usage should be ``r = rand_init(seed, tid)``
   """
   ...

@overload
def randi(state: uint32) -> int:
   """
   Return a random integer between [0, 2^32)
   """
   ...

@overload
def randi(state: uint32, min: int32, max: int32) -> int:
   """
   Return a random integer between [min, max)
   """
   ...

@overload
def randf(state: uint32) -> float:
   """
   Return a random float between [0.0, 1.0)
   """
   ...

@overload
def randf(state: uint32, min: float32, max: float32) -> float:
   """
   Return a random float between [min, max)
   """
   ...

@overload
def randn(state: uint32) -> float:
   """
   Sample a normal distribution
   """
   ...

@overload
def sample_cdf(state: uint32, cdf: array[float32]) -> int:
   """
   Inverse transform sample a cumulative distribution function
   """
   ...

@overload
def sample_triangle(state: uint32) -> vec2:
   """
   Uniformly sample a triangle. Returns sample barycentric coordinates
   """
   ...

@overload
def sample_unit_ring(state: uint32) -> vec2:
   """
   Uniformly sample a ring in the xy plane
   """
   ...

@overload
def sample_unit_disk(state: uint32) -> vec2:
   """
   Uniformly sample a disk in the xy plane
   """
   ...

@overload
def sample_unit_sphere_surface(state: uint32) -> vec3:
   """
   Uniformly sample a unit sphere surface
   """
   ...

@overload
def sample_unit_sphere(state: uint32) -> vec3:
   """
   Uniformly sample a unit sphere
   """
   ...

@overload
def sample_unit_hemisphere_surface(state: uint32) -> vec3:
   """
   Uniformly sample a unit hemisphere surface
   """
   ...

@overload
def sample_unit_hemisphere(state: uint32) -> vec3:
   """
   Uniformly sample a unit hemisphere
   """
   ...

@overload
def sample_unit_square(state: uint32) -> vec2:
   """
   Uniformly sample a unit square
   """
   ...

@overload
def sample_unit_cube(state: uint32) -> vec3:
   """
   Uniformly sample a unit cube
   """
   ...

@overload
def noise(state: uint32, x: float32) -> float:
   """
   Non-periodic Perlin-style noise in 1d.
   """
   ...

@overload
def noise(state: uint32, xy: vec2) -> float:
   """
   Non-periodic Perlin-style noise in 2d.
   """
   ...

@overload
def noise(state: uint32, xyz: vec3) -> float:
   """
   Non-periodic Perlin-style noise in 3d.
   """
   ...

@overload
def noise(state: uint32, xyzt: vec4) -> float:
   """
   Non-periodic Perlin-style noise in 4d.
   """
   ...

@overload
def pnoise(state: uint32, x: float32, px: int32) -> float:
   """
   Periodic Perlin-style noise in 1d.
   """
   ...

@overload
def pnoise(state: uint32, xy: vec2, px: int32, py: int32) -> float:
   """
   Periodic Perlin-style noise in 2d.
   """
   ...

@overload
def pnoise(state: uint32, xyz: vec3, px: int32, py: int32, pz: int32) -> float:
   """
   Periodic Perlin-style noise in 3d.
   """
   ...

@overload
def pnoise(state: uint32, xyzt: vec4, px: int32, py: int32, pz: int32, pt: int32) -> float:
   """
   Periodic Perlin-style noise in 4d.
   """
   ...

@overload
def curlnoise(state: uint32, xy: vec2) -> vec2:
   """
   Divergence-free vector field based on the gradient of a Perlin noise function.
   """
   ...

@overload
def curlnoise(state: uint32, xyz: vec3) -> vec3:
   """
   Divergence-free vector field based on the curl of three Perlin noise functions.
   """
   ...

@overload
def curlnoise(state: uint32, xyzt: vec4) -> vec3:
   """
   Divergence-free vector field based on the curl of three Perlin noise functions.
   """
   ...

@overload
def printf():
   """
   Allows printing formatted strings, using C-style format specifiers.
   """
   ...

@overload
def tid() -> int:
   """
   Return the current thread index. Note that this is the *global* index of the thread in the range [0, dim) 
      where dim is the parameter passed to kernel launch.
   """
   ...

@overload
def tid() -> Tuple[int, int]:
   """
   Return the current thread indices for a 2d kernel launch. Use ``i,j = wp.tid()`` syntax to retrieve the coordinates inside the kernel thread grid.
   """
   ...

@overload
def tid() -> Tuple[int, int, int]:
   """
   Return the current thread indices for a 3d kernel launch. Use ``i,j,k = wp.tid()`` syntax to retrieve the coordinates inside the kernel thread grid.
   """
   ...

@overload
def tid() -> Tuple[int, int, int, int]:
   """
   Return the current thread indices for a 4d kernel launch. Use ``i,j,k,l = wp.tid()`` syntax to retrieve the coordinates inside the kernel thread grid.
   """
   ...

@overload
def select(cond: bool, arg1: Any, arg2: Any):
   """
   Select between two arguments, if cond is false then return ``arg1``, otherwise return ``arg2``
   """
   ...

@overload
def atomic_add(a: array[Any], i: int32, value: Any):
   """
   Atomically add ``value`` onto the array at location given by index.
   """
   ...

@overload
def atomic_add(a: array[Any], i: int32, j: int32, value: Any):
   """
   Atomically add ``value`` onto the array at location given by indices.
   """
   ...

@overload
def atomic_add(a: array[Any], i: int32, j: int32, k: int32, value: Any):
   """
   Atomically add ``value`` onto the array at location given by indices.
   """
   ...

@overload
def atomic_add(a: array[Any], i: int32, j: int32, k: int32, l: int32, value: Any):
   """
   Atomically add ``value`` onto the array at location given by indices.
   """
   ...

@overload
def atomic_sub(a: array[Any], i: int32, value: Any):
   """
   Atomically subtract ``value`` onto the array at location given by index.
   """
   ...

@overload
def atomic_sub(a: array[Any], i: int32, j: int32, value: Any):
   """
   Atomically subtract ``value`` onto the array at location given by indices.
   """
   ...

@overload
def atomic_sub(a: array[Any], i: int32, j: int32, k: int32, value: Any):
   """
   Atomically subtract ``value`` onto the array at location given by indices.
   """
   ...

@overload
def atomic_sub(a: array[Any], i: int32, j: int32, k: int32, l: int32, value: Any):
   """
   Atomically subtract ``value`` onto the array at location given by indices.
   """
   ...

@overload
def atomic_min(a: array[Any], i: int32, value: Any):
   """
   Compute the minimum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.
   """
   ...

@overload
def atomic_min(a: array[Any], i: int32, j: int32, value: Any):
   """
   Compute the minimum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.
   """
   ...

@overload
def atomic_min(a: array[Any], i: int32, j: int32, k: int32, value: Any):
   """
   Compute the minimum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.
   """
   ...

@overload
def atomic_min(a: array[Any], i: int32, j: int32, k: int32, l: int32, value: Any):
   """
   Compute the minimum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.
   """
   ...

@overload
def atomic_max(a: array[Any], i: int32, value: Any):
   """
   Compute the maximum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.
   """
   ...

@overload
def atomic_max(a: array[Any], i: int32, j: int32, value: Any):
   """
   Compute the maximum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.
   """
   ...

@overload
def atomic_max(a: array[Any], i: int32, j: int32, k: int32, value: Any):
   """
   Compute the maximum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.
   """
   ...

@overload
def atomic_max(a: array[Any], i: int32, j: int32, k: int32, l: int32, value: Any):
   """
   Compute the maximum of ``value`` and ``array[index]`` and atomically update the array. Note that for vectors and matrices the operation is only atomic on a per-component basis.
   """
   ...

@overload
def index(a: vec2, i: int32) -> float:
   """

   """
   ...

@overload
def index(a: vec3, i: int32) -> float:
   """

   """
   ...

@overload
def index(a: vec4, i: int32) -> float:
   """

   """
   ...

@overload
def index(a: quat, i: int32) -> float:
   """

   """
   ...

@overload
def index(a: mat22, i: int32) -> vec2:
   """

   """
   ...

@overload
def index(a: mat22, i: int32, j: int32) -> float:
   """

   """
   ...

@overload
def index(a: mat33, i: int32) -> vec3:
   """

   """
   ...

@overload
def index(a: mat33, i: int32, j: int32) -> float:
   """

   """
   ...

@overload
def index(a: mat44, i: int32) -> vec4:
   """

   """
   ...

@overload
def index(a: mat44, i: int32, j: int32) -> float:
   """

   """
   ...

@overload
def index(a: spatial_matrix, i: int32, j: int32) -> float:
   """

   """
   ...

@overload
def index(a: spatial_vector, i: int32) -> float:
   """

   """
   ...

@overload
def index(a: transform, i: int32) -> float:
   """

   """
   ...

@overload
def index(s: shape_t, i: int32) -> int:
   """

   """
   ...

@overload
def expect_eq(arg1: int8, arg2: int8):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: uint8, arg2: uint8):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: int16, arg2: int16):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: uint16, arg2: uint16):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: int32, arg2: int32):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: uint32, arg2: uint32):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: int64, arg2: int64):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: uint64, arg2: uint64):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: float16, arg2: float16):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: float32, arg2: float32):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: float64, arg2: float64):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: vec2, arg2: vec2):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: vec3, arg2: vec3):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: vec4, arg2: vec4):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: mat22, arg2: mat22):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: mat33, arg2: mat33):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: mat44, arg2: mat44):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: quat, arg2: quat):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: transform, arg2: transform):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: spatial_vector, arg2: spatial_vector):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def expect_eq(arg1: spatial_matrix, arg2: spatial_matrix):
   """
   Prints an error to stdout if arg1 and arg2 are not equal
   """
   ...

@overload
def lerp(a: float32, b: float32, t: float32) -> float32:
   """
   Linearly interpolate two values a and b using factor t, computed as ``a*(1-t) + b*t``
   """
   ...

@overload
def lerp(a: vec2, b: vec2, t: float32) -> vec2:
   """
   Linearly interpolate two values a and b using factor t, computed as ``a*(1-t) + b*t``
   """
   ...

@overload
def lerp(a: vec3, b: vec3, t: float32) -> vec3:
   """
   Linearly interpolate two values a and b using factor t, computed as ``a*(1-t) + b*t``
   """
   ...

@overload
def lerp(a: vec4, b: vec4, t: float32) -> vec4:
   """
   Linearly interpolate two values a and b using factor t, computed as ``a*(1-t) + b*t``
   """
   ...

@overload
def lerp(a: mat22, b: mat22, t: float32) -> mat22:
   """
   Linearly interpolate two values a and b using factor t, computed as ``a*(1-t) + b*t``
   """
   ...

@overload
def lerp(a: mat33, b: mat33, t: float32) -> mat33:
   """
   Linearly interpolate two values a and b using factor t, computed as ``a*(1-t) + b*t``
   """
   ...

@overload
def lerp(a: mat44, b: mat44, t: float32) -> mat44:
   """
   Linearly interpolate two values a and b using factor t, computed as ``a*(1-t) + b*t``
   """
   ...

@overload
def lerp(a: quat, b: quat, t: float32) -> quat:
   """
   Linearly interpolate two values a and b using factor t, computed as ``a*(1-t) + b*t``
   """
   ...

@overload
def lerp(a: transform, b: transform, t: float32) -> transform:
   """
   Linearly interpolate two values a and b using factor t, computed as ``a*(1-t) + b*t``
   """
   ...

@overload
def lerp(a: spatial_vector, b: spatial_vector, t: float32) -> spatial_vector:
   """
   Linearly interpolate two values a and b using factor t, computed as ``a*(1-t) + b*t``
   """
   ...

@overload
def lerp(a: spatial_matrix, b: spatial_matrix, t: float32) -> spatial_matrix:
   """
   Linearly interpolate two values a and b using factor t, computed as ``a*(1-t) + b*t``
   """
   ...

@overload
def smoothstep(a: float32, b: float32, t: float32) -> float:
   """
   Smoothly interpolate two values a and b using factor t, using a cubic Hermite interpolation after clamping
   """
   ...

@overload
def expect_near(arg1: float32, arg2: float32, tolerance: float32):
   """
   Prints an error to stdout if arg1 and arg2 are not closer than tolerance in magnitude
   """
   ...

@overload
def expect_near(arg1: vec3, arg2: vec3, tolerance: float32):
   """
   Prints an error to stdout if any element of arg1 and arg2 are not closer than tolerance in magnitude
   """
   ...

@overload
def lower_bound(arr: array[int32], value: int32) -> int:
   """
   Search a sorted array for the closest element greater than or equal to value.
   """
   ...

@overload
def lower_bound(arr: array[float32], value: float32) -> int:
   """
   Search a sorted array for the closest element greater than or equal to value.
   """
   ...

@overload
def add(x: int32, y: int32) -> int:
   """

   """
   ...

@overload
def add(x: float32, y: float32) -> float:
   """

   """
   ...

@overload
def add(x: vec2, y: vec2) -> vec2:
   """

   """
   ...

@overload
def add(x: vec3, y: vec3) -> vec3:
   """

   """
   ...

@overload
def add(x: vec4, y: vec4) -> vec4:
   """

   """
   ...

@overload
def add(x: quat, y: quat) -> quat:
   """

   """
   ...

@overload
def add(x: mat22, y: mat22) -> mat22:
   """

   """
   ...

@overload
def add(x: mat33, y: mat33) -> mat33:
   """

   """
   ...

@overload
def add(x: mat44, y: mat44) -> mat44:
   """

   """
   ...

@overload
def add(x: spatial_vector, y: spatial_vector) -> spatial_vector:
   """

   """
   ...

@overload
def add(x: spatial_matrix, y: spatial_matrix) -> spatial_matrix:
   """

   """
   ...

@overload
def sub(x: int32, y: int32) -> int:
   """

   """
   ...

@overload
def sub(x: float32, y: float32) -> float:
   """

   """
   ...

@overload
def sub(x: vec2, y: vec2) -> vec2:
   """

   """
   ...

@overload
def sub(x: vec3, y: vec3) -> vec3:
   """

   """
   ...

@overload
def sub(x: vec4, y: vec4) -> vec4:
   """

   """
   ...

@overload
def sub(x: mat22, y: mat22) -> mat22:
   """

   """
   ...

@overload
def sub(x: mat33, y: mat33) -> mat33:
   """

   """
   ...

@overload
def sub(x: mat44, y: mat44) -> mat44:
   """

   """
   ...

@overload
def sub(x: spatial_vector, y: spatial_vector) -> spatial_vector:
   """

   """
   ...

@overload
def sub(x: spatial_matrix, y: spatial_matrix) -> spatial_matrix:
   """

   """
   ...

@overload
def mul(x: int32, y: int32) -> int:
   """

   """
   ...

@overload
def mul(x: float32, y: float32) -> float:
   """

   """
   ...

@overload
def mul(x: float32, y: vec2) -> vec2:
   """

   """
   ...

@overload
def mul(x: float32, y: vec3) -> vec3:
   """

   """
   ...

@overload
def mul(x: float32, y: vec4) -> vec4:
   """

   """
   ...

@overload
def mul(x: float32, y: quat) -> quat:
   """

   """
   ...

@overload
def mul(x: float32, y: mat22) -> mat22:
   """

   """
   ...

@overload
def mul(x: float32, y: mat33) -> mat33:
   """

   """
   ...

@overload
def mul(x: float32, y: mat44) -> mat44:
   """

   """
   ...

@overload
def mul(x: vec2, y: float32) -> vec2:
   """

   """
   ...

@overload
def mul(x: vec3, y: float32) -> vec3:
   """

   """
   ...

@overload
def mul(x: vec4, y: float32) -> vec4:
   """

   """
   ...

@overload
def mul(x: quat, y: float32) -> quat:
   """

   """
   ...

@overload
def mul(x: quat, y: quat) -> quat:
   """

   """
   ...

@overload
def mul(x: mat22, y: float32) -> mat22:
   """

   """
   ...

@overload
def mul(x: mat22, y: vec2) -> vec2:
   """

   """
   ...

@overload
def mul(x: mat22, y: mat22) -> mat22:
   """

   """
   ...

@overload
def mul(x: mat33, y: float32) -> mat33:
   """

   """
   ...

@overload
def mul(x: mat33, y: vec3) -> vec3:
   """

   """
   ...

@overload
def mul(x: mat33, y: mat33) -> mat33:
   """

   """
   ...

@overload
def mul(x: mat44, y: float32) -> mat44:
   """

   """
   ...

@overload
def mul(x: mat44, y: vec4) -> vec4:
   """

   """
   ...

@overload
def mul(x: mat44, y: mat44) -> mat44:
   """

   """
   ...

@overload
def mul(x: spatial_vector, y: float32) -> spatial_vector:
   """

   """
   ...

@overload
def mul(x: spatial_matrix, y: spatial_matrix) -> spatial_matrix:
   """

   """
   ...

@overload
def mul(x: spatial_matrix, y: spatial_vector) -> spatial_vector:
   """

   """
   ...

@overload
def mul(x: transform, y: transform) -> transform:
   """

   """
   ...

@overload
def mod(x: int32, y: int32) -> int:
   """

   """
   ...

@overload
def mod(x: float32, y: float32) -> float:
   """

   """
   ...

@overload
def div(x: int32, y: int32) -> int:
   """

   """
   ...

@overload
def div(x: float32, y: float32) -> float:
   """

   """
   ...

@overload
def div(x: vec2, y: float32) -> vec2:
   """

   """
   ...

@overload
def div(x: vec3, y: float32) -> vec3:
   """

   """
   ...

@overload
def div(x: vec4, y: float32) -> vec4:
   """

   """
   ...

@overload
def floordiv(x: int32, y: int32) -> int:
   """

   """
   ...

@overload
def floordiv(x: float32, y: float32) -> float:
   """

   """
   ...

@overload
def neg(x: int32) -> int:
   """

   """
   ...

@overload
def neg(x: float32) -> float:
   """

   """
   ...

@overload
def neg(x: vec2) -> vec2:
   """

   """
   ...

@overload
def neg(x: vec3) -> vec3:
   """

   """
   ...

@overload
def neg(x: vec4) -> vec4:
   """

   """
   ...

@overload
def neg(x: quat) -> quat:
   """

   """
   ...

@overload
def neg(x: mat33) -> mat33:
   """

   """
   ...

@overload
def neg(x: mat44) -> mat44:
   """

   """
   ...

@overload
def unot(b: bool) -> bool:
   """

   """
   ...




# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import math

import warp as wp
import warp.sim

class SimRenderer(warp.render.UsdRenderer):
    
    def __init__(self, model: warp.sim.Model, path):

        from pxr import UsdGeom

        # create USD stage
        super().__init__(path)

        self.model = model

        # add ground plane
        if (self.model.ground):
            self.render_ground(size=20.0)

        # create rigid body root node
        for b in range(model.body_count):
            xform = UsdGeom.Xform.Define(self.stage, self.root.GetPath().AppendChild("body_" + str(b)))
            wp.render._usd_add_xform(xform)

        # create rigid shape children
        if (self.model.shape_count):
            shape_body = model.shape_body.numpy()
            shape_geo_src = model.shape_geo_src#.numpy()
            shape_geo_type = model.shape_geo_type.numpy()
            shape_geo_scale = model.shape_geo_scale.numpy()
            shape_transform = model.shape_transform.numpy()

            for s in range(model.shape_count):
            
                parent_path = self.root.GetPath()
                if shape_body[s] >= 0:
                    parent_path = parent_path.AppendChild("body_" + str(shape_body[s].item()))

                geo_type = shape_geo_type[s]
                geo_scale = shape_geo_scale[s]
                geo_src = shape_geo_src[s]

                # shape transform in body frame
                X_bs = warp.transform_expand(shape_transform[s])

                if (geo_type == warp.sim.GEO_PLANE):

                    # plane mesh
                    size = 1000.0

                    mesh = UsdGeom.Mesh.Define(self.stage, parent_path.AppendChild("plane_" + str(s)))
                    mesh.CreateDoubleSidedAttr().Set(True)

                    points = ((-size, 0.0, -size), (size, 0.0, -size), (size, 0.0, size), (-size, 0.0, size))
                    normals = ((0.0, 1.0, 0.0), (0.0, 1.0, 0.0), (0.0, 1.0, 0.0), (0.0, 1.0, 0.0))
                    counts = (4, )
                    indices = [0, 1, 2, 3]

                    mesh.GetPointsAttr().Set(points)
                    mesh.GetNormalsAttr().Set(normals)
                    mesh.GetFaceVertexCountsAttr().Set(counts)
                    mesh.GetFaceVertexIndicesAttr().Set(indices)

                elif (geo_type == warp.sim.GEO_SPHERE):

                    mesh = UsdGeom.Sphere.Define(self.stage, parent_path.AppendChild("sphere_" + str(s)))
                    mesh.GetRadiusAttr().Set(float(geo_scale[0]))

                    wp.render._usd_add_xform(mesh)
                    wp.render._usd_set_xform(mesh, X_bs.p, X_bs.q, (1.0, 1.0, 1.0), 0.0)

                elif (geo_type == warp.sim.GEO_CAPSULE):
                    mesh = UsdGeom.Capsule.Define(self.stage, parent_path.AppendChild("capsule_" + str(s)))
                    mesh.GetRadiusAttr().Set(float(geo_scale[0]))
                    mesh.GetHeightAttr().Set(float(geo_scale[1] * 2.0))

                    # geometry transform w.r.t shape, convert USD geometry to physics engine convention
                    X_sg = warp.transform((0.0, 0.0, 0.0), warp.utils.quat_from_axis_angle((0.0, 1.0, 0.0), math.pi * 0.5))
                    X_bg = warp.utils.transform_multiply(X_bs, X_sg)

                    wp.render._usd_add_xform(mesh)
                    wp.render._usd_set_xform(mesh, X_bg.p, X_bg.q, (1.0, 1.0, 1.0), 0.0)

                elif (geo_type == warp.sim.GEO_BOX):
                    mesh = UsdGeom.Cube.Define(self.stage, parent_path.AppendChild("box_" + str(s)))
                    #mesh.GetSizeAttr().Set((geo_scale[0], geo_scale[1], geo_scale[2]))

                    wp.render._usd_add_xform(mesh)
                    wp.render._usd_set_xform(mesh, X_bs.p, X_bs.q, (geo_scale[0], geo_scale[1], geo_scale[2]), 0.0)

                elif (geo_type == warp.sim.GEO_MESH):

                    mesh = UsdGeom.Mesh.Define(self.stage, parent_path.AppendChild("mesh_" + str(s)))
                    mesh.GetPointsAttr().Set(geo_src.vertices)
                    mesh.GetFaceVertexIndicesAttr().Set(geo_src.indices)
                    mesh.GetFaceVertexCountsAttr().Set([3] * int(len(geo_src.indices) / 3))

                    wp.render._usd_add_xform(mesh)
                    wp.render._usd_set_xform(mesh, X_bs.p, X_bs.q, (geo_scale[0], geo_scale[1], geo_scale[2]), 0.0)

                elif (geo_type == warp.sim.GEO_SDF):
                    pass
        


    def render(self, state: warp.sim.State):

        from pxr import UsdGeom, Gf, Sdf

        if (self.model.particle_count):

            particle_q = state.particle_q.numpy()

            # render particles
            self.render_points("particles", particle_q, radius=0.1)

            # render tris
            if (self.model.tri_count):
                self.render_mesh("surface", particle_q, self.model.tri_indices.numpy().flatten())

            # render springs
            if (self.model.spring_count):
                self.render_line_list("springs", particle_q, self.model.spring_indices.numpy().flatten(), [], 0.1)

        # render muscles
        if (self.model.muscle_count):
            
            body_q = state.body_q.numpy()

            muscle_start = self.model.muscle_start.numpy()
            muscle_links = self.model.muscle_bodies.numpy()
            muscle_points = self.model.muscle_points.numpy()
            muscle_activation = self.model.muscle_activation.numpy()

            # for s in self.skeletons:
                
            #     # for mesh, link in s.mesh_map.items():
                    
            #     #     if link != -1:
            #     #         X_sc = wp.transform_expand(self.state.body_X_sc[link].tolist())

            #     #         #self.renderer.add_mesh(mesh, "../assets/snu/OBJ/" + mesh + ".usd", X_sc, 1.0, self.render_time)
            #     #         self.renderer.add_mesh(mesh, "../assets/snu/OBJ/" + mesh + ".usd", X_sc, 1.0, self.render_time)

            for m in range(self.model.muscle_count):

                start = int(muscle_start[m])
                end = int(muscle_start[m + 1])

                points = []

                for w in range(start, end):
                    
                    link = muscle_links[w]
                    point = muscle_points[w]

                    X_sc = wp.transform_expand(body_q[link][0])

                    points.append(Gf.Vec3f(wp.transform_point(X_sc, point).tolist()))
                
                self.render_line_strip(name=f"muscle_{m}", vertices=points, radius=0.0075, color=(muscle_activation[m], 0.2, 0.5))
        
        
        with Sdf.ChangeBlock():

            # update  bodies
            if (self.model.body_count):

                body_q = state.body_q.numpy()

                for b in range(self.model.body_count):

                    node = UsdGeom.Xform(self.stage.GetPrimAtPath(self.root.GetPath().AppendChild("body_" + str(b))))

                    # unpack rigid transform
                    X_sb = warp.transform_expand(body_q[b])

                    wp.render._usd_set_xform(node, X_sb.p, X_sb.q, (1.0, 1.0, 1.0), self.time)





# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from dataclasses import dataclass
from typing import Any
import unittest

import numpy as np

import warp as wp
from warp.tests.test_base import *

@dataclass
class TestData:
    a: Any
    b: Any
    t: float
    expected: Any
    expected_adj_a: Any = None
    expected_adj_b: Any = None
    expected_adj_t: float = None

    def check_backwards(self):
        return (
            self.expected_adj_a is not None
            and self.expected_adj_b is not None
            and self.expected_adj_t is not None
        )

TEST_DATA = {
    wp.float32: (
        TestData(
            a=1.0,
            b=5.0,
            t=0.75,
            expected=4.0,
            expected_adj_a=0.25,
            expected_adj_b=0.75,
            expected_adj_t=4.0,
        ),
        TestData(
            a=-2.0,
            b=5.0,
            t=0.25,
            expected=-0.25,
            expected_adj_a=0.75,
            expected_adj_b=0.25,
            expected_adj_t=7.0,
        ),
        TestData(
            a=1.23,
            b=2.34,
            t=0.5,
            expected=1.785,
            expected_adj_a=0.5,
            expected_adj_b=0.5,
            expected_adj_t=1.11,
        ),
    ),
    wp.vec2: (
        TestData(
            a=[1, 2],
            b=[3, 4],
            t=0.5,
            expected=[2, 3],
        ),
    ),
    wp.vec3: (
        TestData(
            a=[1, 2, 3],
            b=[3, 4, 5],
            t=0.5,
            expected=[2, 3, 4],
        ),
    ),
    wp.vec4: (
        TestData(
            a=[1, 2, 3, 4],
            b=[3, 4, 5, 6],
            t=0.5,
            expected=[2, 3, 4, 5],
        ),
    ),
    wp.mat22: (
        TestData(
            a=[[1, 2], [2, 1]],
            b=[[3, 4], [4, 3]],
            t=0.5,
            expected=[[2, 3], [3, 2]],
        ),
    ),
    wp.mat33: (
        TestData(
            a=[[1, 2, 3], [3, 1, 2], [2, 3, 1]],
            b=[[3, 4, 5], [5, 3, 4], [4, 5, 3]],
            t=0.5,
            expected=[[2, 3, 4], [4, 2, 3], [3, 4, 2]],
        ),
    ),
    wp.mat44: (
        TestData(
            a=[[1, 2, 3, 4], [4, 1, 2, 3], [3, 4, 1, 2], [2, 3, 4, 1]],
            b=[[3, 4, 5, 6], [6, 3, 4, 5], [5, 6, 3, 4], [4, 5, 6, 3]],
            t=0.5,
            expected=[[2, 3, 4, 5], [5, 2, 3, 4], [4, 5, 2, 3], [3, 4, 5, 2]],
        ),
    ),
    wp.quat: (
        TestData(
            a=[1, 2, 3, 4],
            b=[3, 4, 5, 6],
            t=0.5,
            expected=[2, 3, 4, 5],
        ),
    ),
    wp.transform: (
        TestData(
            a=[1, 2, 3, 4, 5, 6, 7],
            b=[3, 4, 5, 6, 7, 8, 9],
            t=0.5,
            expected=[2, 3, 4, 5, 6, 7, 8],
        ),
    ),
    wp.spatial_vector: (
        TestData(
            a=[1, 2, 3, 4, 5, 6],
            b=[3, 4, 5, 6, 7, 8],
            t=0.5,
            expected=[2, 3, 4, 5, 6, 7],
        ),
    ),
    wp.spatial_matrix: (
        TestData(
            a=[[1, 2, 3, 4, 5, 6],
               [6, 1, 2, 3, 4, 5],
               [5, 6, 1, 2, 3, 4],
               [4, 5, 6, 1, 2, 3],
               [3, 4, 5, 6, 1, 2],
               [2, 3, 4, 5, 6, 1]],
            b=[[3, 4, 5, 6, 7, 8],
               [8, 3, 4, 5, 6, 7],
               [7, 8, 3, 4, 5, 6],
               [6, 7, 8, 3, 4, 5],
               [5, 6, 7, 8, 3, 4],
               [4, 5, 6, 7, 8, 3]],
            t=0.5,
            expected=[[2, 3, 4, 5, 6, 7],
                      [7, 2, 3, 4, 5, 6],
                      [6, 7, 2, 3, 4, 5],
                      [5, 6, 7, 2, 3, 4],
                      [4, 5, 6, 7, 2, 3],
                      [3, 4, 5, 6, 7, 2]],
        ),
    ),
}

wp.init()

def test_lerp(test, device):

    def make_kernel_fn(data_type):

        def fn(
            a: wp.array(dtype=data_type),
            b: wp.array(dtype=data_type),
            t: wp.array(dtype=float),
            out: wp.array(dtype=data_type),
        ):
            out[0] = wp.lerp(a[0], b[0], t[0])

        return fn

    for data_type in TEST_DATA:
        kernel_fn = make_kernel_fn(data_type)
        module = wp.get_module(kernel_fn.__module__)
        kernel = wp.Kernel(
            func=kernel_fn,
            key=f"test_lerp_{data_type.__name__}_kernel",
            module=module,
        )

        for test_data in TEST_DATA[data_type]:
            a = wp.array(
                [test_data.a],
                dtype=data_type,
                device=device,
                requires_grad=True,
            )
            b = wp.array(
                [test_data.b],
                dtype=data_type,
                device=device,
                requires_grad=True,
            )
            t = wp.array(
                [test_data.t],
                dtype=float,
                device=device,
                requires_grad=True,
            )
            out = wp.array(
                [0] * wp.types.type_length(data_type),
                dtype=data_type,
                device=device,
                requires_grad=True,
            )

            tape = wp.Tape()
            with tape:
                wp.launch(
                    kernel,
                    dim=1,
                    inputs=[a, b, t, out],
                    device=device,
                )

            assert_np_equal(
                out.numpy(),
                np.array([test_data.expected]),
                tol=1e-6,
            )

            if test_data.check_backwards():
                tape.backward(out)

                assert_np_equal(
                    tape.gradients[a].numpy(),
                    np.array([test_data.expected_adj_a]),
                    tol=1e-6,
                )
                assert_np_equal(
                    tape.gradients[b].numpy(),
                    np.array([test_data.expected_adj_b]),
                    tol=1e-6,
                )
                assert_np_equal(
                    tape.gradients[t].numpy(),
                    np.array([test_data.expected_adj_t]),
                    tol=1e-6,
                )

def register(parent):
    devices = wp.get_devices()

    class TestLerp(parent):
        pass

    add_function_test(TestLerp, "test_lerp", test_lerp, devices=devices)
    return TestLerp

if __name__ == "__main__":
    _ = register(unittest.TestCase)
    unittest.main(verbosity=2)



# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import numpy as np
import math

import warp as wp
from warp.tests.test_base import *

wp.init()

@wp.kernel
def test_operators_scalar_float():

    a = 1.0
    b = 2.0

    c = a*b
    d = a+b
    e = a/b
    f = a-b
    g = b**8.0
    h = 10.0//3.0

    expect_eq(c, 2.0)
    expect_eq(d, 3.0)
    expect_eq(e, 0.5)
    expect_eq(f, -1.0)
    expect_eq(g, 256.0)
    expect_eq(h, 3.0)
    
@wp.kernel
def test_operators_scalar_int():

    a = 1
    b = 2

    c = a*b
    d = a+b
    e = a/b
    f = a-b
    # g = b**8    # integer pow not implemented
    h = 10//3

    expect_eq(c, 2)
    expect_eq(d, 3)
    expect_eq(e, 0)
    expect_eq(f, -1)
    # expect_eq(g, 256)
    expect_eq(h, 3)

@wp.kernel
def test_operators_vector_index():

    v = wp.vec4(1.0, 2.0, 3.0, 4.0)

    expect_eq(v[0], 1.0)
    expect_eq(v[1], 2.0)
    expect_eq(v[2], 3.0)
    expect_eq(v[3], 4.0)


@wp.kernel
def test_operators_matrix_index():

    m22 = wp.mat22(1.0, 2.0, 3.0, 4.0)

    expect_eq(m22[0,0], 1.0)
    expect_eq(m22[0,1], 2.0)
    expect_eq(m22[1,0], 3.0)
    expect_eq(m22[1,1], 4.0)
    


@wp.kernel
def test_operators_vec3():
    
    v = vec3(1.0, 2.0, 3.0)

    r0 = v*3.0
    r1 = 3.0*v

    expect_eq(r0, vec3(3.0, 6.0, 9.0))
    expect_eq(r1, vec3(3.0, 6.0, 9.0))

    col0 = vec3(1.0, 0.0, 0.0)
    col1 = vec3(0.0, 2.0, 0.0)
    col2 = vec3(0.0, 0.0, 3.0)

    m = mat33(col0, col1, col2)

    expect_eq(m*vec3(1.0, 0.0, 0.0), col0)
    expect_eq(m*vec3(0.0, 1.0, 0.0), col1)
    expect_eq(m*vec3(0.0, 0.0, 1.0), col2)

    two = vec3(1.0)*2.0
    expect_eq(two, vec3(2.0, 2.0, 2.0))

@wp.kernel
def test_operators_vec4():
    
    v = vec4(1.0, 2.0, 3.0, 4.0)

    r0 = v*3.0
    r1 = 3.0*v

    expect_eq(r0, vec4(3.0, 6.0, 9.0, 12.0))
    expect_eq(r1, vec4(3.0, 6.0, 9.0, 12.0))

    col0 = vec4(1.0, 0.0, 0.0, 0.0)
    col1 = vec4(0.0, 2.0, 0.0, 0.0)
    col2 = vec4(0.0, 0.0, 3.0, 0.0)
    col3 = vec4(0.0, 0.0, 0.0, 4.0)

    m = mat44(col0, col1, col2, col3)

    expect_eq(m*vec4(1.0, 0.0, 0.0, 0.0), col0)
    expect_eq(m*vec4(0.0, 1.0, 0.0, 0.0), col1)
    expect_eq(m*vec4(0.0, 0.0, 1.0, 0.0), col2)
    expect_eq(m*vec4(0.0, 0.0, 0.0, 1.0), col3)

    two = vec4(1.0)*2.0
    expect_eq(two, vec4(2.0, 2.0, 2.0, 2.0))
       


@wp.kernel
def test_operators_mat22():
    
    m = mat22(1.0, 2.0, 3.0, 4.0)
    r = mat22(3.0, 6.0, 9.0, 12.0)

    r0 = m*3.0
    r1 = 3.0*m

    expect_eq(r0, r)
    expect_eq(r1, r)

    expect_eq(r0[0,0], 3.0)
    expect_eq(r0[0,1], 6.0)
    expect_eq(r0[1,0], 9.0)
    expect_eq(r0[1,1], 12.0)

    expect_eq(r0[0], wp.vec2(3.0, 6.0))
    expect_eq(r0[1], wp.vec2(9.0, 12.0))

@wp.kernel
def test_operators_mat33():
    
    m = mat33(1.0, 2.0, 3.0,
              4.0, 5.0, 6.0,
              7.0, 8.0, 9.0)

    r = mat33(3.0, 6.0, 9.0,
              12.0, 15.0, 18.0,
              21.0, 24.0, 27.0)              

    r0 = m*3.0
    r1 = 3.0*m

    expect_eq(r0, r)
    expect_eq(r1, r)

    expect_eq(r0[0,0], 3.0)
    expect_eq(r0[0,1], 6.0)
    expect_eq(r0[0,2], 9.0)

    expect_eq(r0[1,0], 12.0)
    expect_eq(r0[1,1], 15.0)
    expect_eq(r0[1,2], 18.0)

    expect_eq(r0[2,0], 21.0)
    expect_eq(r0[2,1], 24.0)
    expect_eq(r0[2,2], 27.0)

    expect_eq(r0[0], wp.vec3(3.0, 6.0, 9.0))
    expect_eq(r0[1], wp.vec3(12.0, 15.0, 18.0))
    expect_eq(r0[2], wp.vec3(21.0, 24.0, 27.0))



@wp.kernel
def test_operators_mat44():
    
    m = mat44(1.0, 2.0, 3.0, 4.0,
              5.0, 6.0, 7.0, 8.0,
              9.0, 10.0, 11.0, 12.0,
              13.0, 14.0, 15.0, 16.0)

    r = mat44(3.0, 6.0, 9.0, 12.0,
              15.0, 18.0, 21.0, 24.0,
              27.0, 30.0, 33.0, 36.0,
              39.0, 42.0, 45.0, 48.0)

    r0 = m*3.0
    r1 = 3.0*m

    expect_eq(r0, r)
    expect_eq(r1, r)

    expect_eq(r0[0,0], 3.0)
    expect_eq(r0[0,1], 6.0)
    expect_eq(r0[0,2], 9.0)
    expect_eq(r0[0,3], 12.0)

    expect_eq(r0[1,0], 15.0)
    expect_eq(r0[1,1], 18.0)
    expect_eq(r0[1,2], 21.0)
    expect_eq(r0[1,3], 24.0)

    expect_eq(r0[2,0], 27.0)
    expect_eq(r0[2,1], 30.0)
    expect_eq(r0[2,2], 33.0)
    expect_eq(r0[2,3], 36.0)

    expect_eq(r0[3,0], 39.0)
    expect_eq(r0[3,1], 42.0)
    expect_eq(r0[3,2], 45.0)
    expect_eq(r0[3,3], 48.0)

    expect_eq(r0[0], wp.vec4(3.0, 6.0, 9.0, 12.0))
    expect_eq(r0[1], wp.vec4(15.0, 18.0, 21.0, 24.0))
    expect_eq(r0[2], wp.vec4(27.0, 30.0, 33.0, 36.0))
    expect_eq(r0[3], wp.vec4(39.0, 42.0, 45.0, 48.0))



def register(parent):

    devices = wp.get_devices()

    class TestOperators(parent):
        pass

    add_kernel_test(TestOperators, test_operators_scalar_float, dim=1, devices=devices)
    add_kernel_test(TestOperators, test_operators_scalar_int, dim=1, devices=devices)
    add_kernel_test(TestOperators, test_operators_matrix_index, dim=1, devices=devices)
    add_kernel_test(TestOperators, test_operators_vector_index, dim=1, devices=devices)
    add_kernel_test(TestOperators, test_operators_vec3, dim=1, devices=devices)
    add_kernel_test(TestOperators, test_operators_vec4, dim=1, devices=devices)

    add_kernel_test(TestOperators, test_operators_mat22, dim=1, devices=devices)
    add_kernel_test(TestOperators, test_operators_mat33, dim=1, devices=devices)
    add_kernel_test(TestOperators, test_operators_mat44, dim=1, devices=devices)

    return TestOperators

if __name__ == '__main__':
    wp.force_load()

    c = register(unittest.TestCase)
    unittest.main(verbosity=2)



# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import warp as wp
from warp.tests.test_base import *
from warp.sim import ModelBuilder

import numpy as np

wp.init()

devices = wp.get_devices()

def register(parent):

    class TestModel(parent):
        def test_add_triangles(self):
            pts = np.random.randn(5,3)
            tris = np.array([[0,3,1],[1,3,4],[1,4,2],[1,4,1]])
            
            builder1 = ModelBuilder()
            builder2 = ModelBuilder()
            for pt in pts:
                builder1.add_particle(pt, [0.0,0.0,0.0], 1.0)
                builder2.add_particle(pt, [0.0,0.0,0.0], 1.0)
            
            # test add_triangle(s) with default arguments:
            for t in tris:
                builder1.add_triangle(t[0],t[1],t[2])
            builder2.add_triangles( tris[:,0], tris[:,1], tris[:,2])
            
            # test add_triangle(s) with non default arguments:
            tri_ke = np.random.randn(3)
            tri_ka = np.random.randn(3)
            tri_kd = np.random.randn(3)
            tri_drag = np.random.randn(3)
            tri_lift = np.random.randn(3)
            for i in range(3):
                t = tris[i]
                builder1.add_triangle(t[0],t[1],t[2],tri_ke[i],tri_ka[i],tri_kd[i],tri_drag[i],tri_lift[i],)
            builder2.add_triangles( tris[:,0], tris[:,1], tris[:,2], tri_ke, tri_ka, tri_kd, tri_drag, tri_lift )

            assert_np_equal(np.array(builder1.tri_indices), np.array(builder2.tri_indices))
            assert_np_equal(np.array(builder1.tri_poses), np.array(builder2.tri_poses), tol=1.e-6)
            assert_np_equal(np.array(builder1.tri_activations), np.array(builder2.tri_activations))
            assert_np_equal(np.array(builder1.tri_materials), np.array(builder2.tri_materials))

            
        def test_add_edges(self):
            pts = np.random.randn(5,3)
            edges = np.array([[0,4,3,1],[3,2,4,1]])
            
            builder1 = ModelBuilder()
            builder2 = ModelBuilder()
            for pt in pts:
                builder1.add_particle(pt, [0.0,0.0,0.0], 1.0)
                builder2.add_particle(pt, [0.0,0.0,0.0], 1.0)
            
            # test defaults:
            for i in range(2):
                builder1.add_edge(edges[i,0],edges[i,1],edges[i,2],edges[i,3])
            builder2.add_edges(edges[:,0],edges[:,1],edges[:,2],edges[:,3])

            # test non defaults:
            rest = np.random.randn(2)
            edge_ke = np.random.randn(2)
            edge_kd = np.random.randn(2)
            for i in range(2):
                builder1.add_edge(edges[i,0],edges[i,1],edges[i,2],edges[i,3],rest[i],edge_ke[i],edge_kd[i])
            builder2.add_edges(edges[:,0],edges[:,1],edges[:,2],edges[:,3],rest,edge_ke,edge_kd)

            assert_np_equal(np.array(builder1.edge_indices), np.array(builder2.edge_indices))
            assert_np_equal(np.array(builder1.edge_rest_angle), np.array(builder2.edge_rest_angle), tol=1.e-4)
            assert_np_equal(np.array(builder1.edge_bending_properties), np.array(builder2.edge_bending_properties))

            

    return TestModel

if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2)



import warp as wp
import numpy as np

import unittest

import warp as wp
from warp.tests.test_base import *


wp.init()

@wp.kernel
def load_store_half(f32: wp.array(dtype=wp.float32),
            f16: wp.array(dtype=wp.float16)):

    tid = wp.tid()

    # check conversion from f32->f16
    a = wp.float16(f32[tid])
    b = f16[tid]
    
    wp.expect_eq(a, b)

    # check stores
    f16[tid] = a


def test_fp16_conversion(test, device):
        
    s = [1.0, 2.0, 3.0, -3.14159]

    np_f32 = np.array(s, dtype=np.float32)
    np_f16 = np.array(s, dtype=np.float16)

    wp_f32 = wp.array(s, dtype=wp.float32, device=device)
    wp_f16 = wp.array(s, dtype=wp.float16, device=device)

    assert_np_equal(np_f32, wp_f32.numpy())
    assert_np_equal(np_f16, wp_f16.numpy())

    wp.launch(load_store_half, dim=len(s), inputs=[wp_f32, wp_f16], device=device)

    # check that stores worked
    assert_np_equal(np_f16, wp_f16.numpy())


@wp.kernel
def mul_half(input: wp.array(dtype=wp.float16),
             output: wp.array(dtype=wp.float16)):

    tid = wp.tid()

    # convert to compute type fp32
    x = wp.float(input[tid])*2.0

    # store back as fp16
    output[tid] = wp.float16(x)


def test_fp16_grad(test, device):
    # checks that gradients are correctly propagated for
    # fp16 arrays, even when intermediate calcualtions
    # are performed in e.g.: fp32

    s = np.random.rand(15).astype(np.float16)

    input = wp.array(s, dtype=wp.float16, device=device, requires_grad=True)
    output = wp.zeros_like(input)

    tape = wp.Tape()
    with tape:
        wp.launch(mul_half, dim=len(s), inputs=[input, output], device=device)

    ones = wp.array(np.ones(len(output)), dtype=wp.float16, device=device)

    tape.backward(grads={output: ones})
    
    assert_np_equal(input.grad.numpy(), np.ones(len(s))*2.0)


def register(parent):

    class TestFp16(parent):
        pass
    
    devices = []
    if wp.is_cpu_available():
        devices.append("cpu")
    for cuda_device in wp.get_cuda_devices():
        if cuda_device.arch >= 70:
            devices.append(cuda_device)

    add_function_test(TestFp16, "test_fp16_conversion", test_fp16_conversion, devices=devices)
    add_function_test(TestFp16, "test_fp16_grad", test_fp16_grad, devices=devices)

    return TestFp16


if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2)





# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import warp as wp


@wp.kernel
def integrate_particles(x: wp.array(dtype=wp.vec3),
                        v: wp.array(dtype=wp.vec3),
                        f: wp.array(dtype=wp.vec3),
                        w: wp.array(dtype=float),
                        gravity: wp.vec3,
                        dt: float,
                        x_new: wp.array(dtype=wp.vec3),
                        v_new: wp.array(dtype=wp.vec3)):

    tid = wp.tid()

    x0 = x[tid]
    v0 = v[tid]
    f0 = f[tid]
    inv_mass = w[tid]

    # simple semi-implicit Euler. v1 = v0 + a dt, x1 = x0 + v1 dt
    v1 = v0 + (f0 * inv_mass + gravity * wp.step(0.0 - inv_mass)) * dt
    x1 = x0 + v1 * dt

    x_new[tid] = x1
    v_new[tid] = v1


@wp.kernel
def solve_springs(x: wp.array(dtype=wp.vec3),
                 v: wp.array(dtype=wp.vec3),
                 invmass: wp.array(dtype=float),
                 spring_indices: wp.array(dtype=int),
                 spring_rest_lengths: wp.array(dtype=float),
                 spring_stiffness: wp.array(dtype=float),
                 spring_damping: wp.array(dtype=float),
                 dt: float,
                 delta: wp.array(dtype=wp.vec3)):

    tid = wp.tid()

    i = spring_indices[tid * 2 + 0]
    j = spring_indices[tid * 2 + 1]

    ke = spring_stiffness[tid]
    kd = spring_damping[tid]
    rest = spring_rest_lengths[tid]

    xi = x[i]
    xj = x[j]

    vi = v[i]
    vj = v[j]

    xij = xi - xj
    vij = vi - vj

    l = length(xij)
    l_inv = 1.0 / l

    # normalized spring direction
    dir = xij * l_inv

    c = l - rest
    dcdt = dot(dir, vij)

    # damping based on relative velocity.
    #fs = dir * (ke * c + kd * dcdt)

    wi = invmass[i]
    wj = invmass[j]

    denom = wi + wj
    alpha = 1.0/(ke*dt*dt)

    multiplier = c / (denom)# + alpha)

    xd = dir*multiplier

    wp.atomic_sub(delta, i, xd*wi)
    wp.atomic_add(delta, j, xd*wj)



@wp.kernel
def solve_tetrahedra(x: wp.array(dtype=wp.vec3),
                     v: wp.array(dtype=wp.vec3),
                     inv_mass: wp.array(dtype=float),
                     indices: wp.array(dtype=int),
                     pose: wp.array(dtype=wp.mat33),
                     activation: wp.array(dtype=float),
                     materials: wp.array(dtype=float),
                     dt: float,
                     relaxation: float,
                     delta: wp.array(dtype=wp.vec3)):

    tid = wp.tid()

    i = indices[tid * 4 + 0]
    j = indices[tid * 4 + 1]
    k = indices[tid * 4 + 2]
    l = indices[tid * 4 + 3]

    act = activation[tid]

    k_mu = materials[tid * 3 + 0]
    k_lambda = materials[tid * 3 + 1]
    k_damp = materials[tid * 3 + 2]

    x0 = x[i]
    x1 = x[j]
    x2 = x[k]
    x3 = x[l]

    v0 = v[i]
    v1 = v[j]
    v2 = v[k]
    v3 = v[l]

    w0 = inv_mass[i]
    w1 = inv_mass[j]
    w2 = inv_mass[k]
    w3 = inv_mass[l]

    x10 = x1 - x0
    x20 = x2 - x0
    x30 = x3 - x0

    v10 = v1 - v0
    v20 = v2 - v0
    v30 = v3 - v0

    Ds = wp.mat33(x10, x20, x30)
    Dm = pose[tid]

    inv_rest_volume = wp.determinant(Dm) * 6.0
    rest_volume = 1.0 / inv_rest_volume

    # F = Xs*Xm^-1
    F = Ds * Dm

    f1 = wp.vec3(F[0, 0], F[1, 0], F[2, 0])
    f2 = wp.vec3(F[0, 1], F[1, 1], F[2, 1])
    f3 = wp.vec3(F[0, 2], F[1, 2], F[2, 2])

    # # C_sqrt
    # tr = dot(f1, f1) + dot(f2, f2) + dot(f3, f3)
    # r_s = wp.sqrt(abs(tr - 3.0))
    # C = r_s

    # if (r_s == 0.0):
    #     return
    
    # if (tr < 3.0):
    #     r_s = 0.0 - r_s

    # dCdx = F*wp.transpose(Dm)*(1.0/r_s)
    # alpha = 1.0 + k_mu / k_lambda

    # C_Neo
    r_s = wp.sqrt(dot(f1, f1) + dot(f2, f2) + dot(f3, f3))
    r_s_inv = 1.0/r_s
    C = r_s
    dCdx = F*wp.transpose(Dm)*r_s_inv
    alpha = 1.0 + k_mu / k_lambda

    # C_Spherical
    # r_s = wp.sqrt(dot(f1, f1) + dot(f2, f2) + dot(f3, f3))
    # r_s_inv = 1.0/r_s
    # C = r_s - wp.sqrt(3.0) 
    # dCdx = F*wp.transpose(Dm)*r_s_inv
    # alpha = 1.0

    # C_D
    #r_s = wp.sqrt(dot(f1, f1) + dot(f2, f2) + dot(f3, f3))
    #C = r_s*r_s - 3.0
    #dCdx = F*wp.transpose(Dm)*2.0
    #alpha = 1.0

    grad1 = vec3(dCdx[0,0], dCdx[1,0], dCdx[2,0])
    grad2 = vec3(dCdx[0,1], dCdx[1,1], dCdx[2,1])
    grad3 = vec3(dCdx[0,2], dCdx[1,2], dCdx[2,2])
    grad0 = (grad1 + grad2 + grad3)*(0.0 - 1.0)

    denom = dot(grad0,grad0)*w0 + dot(grad1,grad1)*w1 + dot(grad2,grad2)*w2 + dot(grad3,grad3)*w3
    multiplier = C/(denom + 1.0/(k_mu*dt*dt*rest_volume))

    delta0 = grad0*multiplier
    delta1 = grad1*multiplier
    delta2 = grad2*multiplier
    delta3 = grad3*multiplier



    # hydrostatic part
    J = wp.determinant(F)


    C_vol = J - alpha
    # dCdx = wp.mat33(cross(f2, f3), cross(f3, f1), cross(f1, f2))*wp.transpose(Dm)

    # grad1 = float3(dCdx[0,0], dCdx[1,0], dCdx[2,0])
    # grad2 = float3(dCdx[0,1], dCdx[1,1], dCdx[2,1])
    # grad3 = float3(dCdx[0,2], dCdx[1,2], dCdx[2,2])
    # grad0 = (grad1 + grad2 + grad3)*(0.0 - 1.0)

    s = inv_rest_volume / 6.0
    grad1 = wp.cross(x20, x30) * s
    grad2 = wp.cross(x30, x10) * s
    grad3 = wp.cross(x10, x20) * s
    grad0 = (grad1 + grad2 + grad3)*(0.0 - 1.0)

    denom = dot(grad0, grad0)*w0 + dot(grad1, grad1)*w1 + dot(grad2, grad2)*w2 + dot(grad3, grad3)*w3
    multiplier = C_vol/(denom + 1.0/(k_lambda*dt*dt*rest_volume))

    delta0 = delta0 + grad0 * multiplier
    delta1 = delta1 + grad1 * multiplier
    delta2 = delta2 + grad2 * multiplier
    delta3 = delta3 + grad3 * multiplier

    # apply forces
    wp.atomic_sub(delta, i, delta0*w0*relaxation)
    wp.atomic_sub(delta, j, delta1*w1*relaxation)
    wp.atomic_sub(delta, k, delta2*w2*relaxation)
    wp.atomic_sub(delta, l, delta3*w3*relaxation)


@wp.kernel
def apply_deltas(x_orig: wp.array(dtype=wp.vec3),
                 v_orig: wp.array(dtype=wp.vec3),
                 x_pred: wp.array(dtype=wp.vec3),
                 delta: wp.array(dtype=wp.vec3),
                 dt: float,
                 x_out: wp.array(dtype=wp.vec3),
                 v_out: wp.array(dtype=wp.vec3)):

    tid = wp.tid()

    x0 = x_orig[tid]
    xp = x_pred[tid]

    # constraint deltas
    d = delta[tid]

    x_new = xp + d
    v_new = (x_new - x0)/dt

    x_out[tid] = x_new
    v_out[tid] = v_new

    # clear forces
    delta[tid] = wp.vec3()


class XPBDIntegrator:
    """A implicit integrator using XPBD

    After constructing `Model` and `State` objects this time-integrator
    may be used to advance the simulation state forward in time.

    Semi-implicit time integration is a variational integrator that 
    preserves energy, however it not unconditionally stable, and requires a time-step
    small enough to support the required stiffness and damping forces.

    See: https://en.wikipedia.org/wiki/Semi-implicit_Euler_method

    Example:

        >>> integrator = wp.SemiImplicitIntegrator()
        >>>
        >>> # simulation loop
        >>> for i in range(100):
        >>>     state = integrator.forward(model, state, dt)

    """

    def __init__(self, iterations, relaxation):
        
        self.iterations = iterations
        self.relaxation = relaxation


    def simulate(self, model, state_in, state_out, dt):

        with wp.ScopedTimer("simulate", False):

            q_pred = wp.zeros_like(state_in.particle_q)
            qd_pred = wp.zeros_like(state_in.particle_qd)

            # alloc particle force buffer
            if (model.particle_count):
                state_out.particle_f.zero_()


            #----------------------------
            # integrate particles

            if (model.particle_count):
                wp.launch(kernel=integrate_particles,
                            dim=model.particle_count,
                            inputs=[state_in.particle_q, state_in.particle_qd, state_out.particle_f, model.particle_inv_mass, model.gravity, dt],
                            outputs=[q_pred, qd_pred],
                            device=model.device)


            for i in range(self.iterations):

                # damped springs
                if (model.spring_count):

                    wp.launch(kernel=solve_springs,
                                dim=model.spring_count,
                                inputs=[state_in.particle_q, state_in.particle_qd, model.particle_inv_mass, model.spring_indices, model.spring_rest_length, model.spring_stiffness, model.spring_damping, dt],
                                outputs=[state_out.particle_f],
                                device=model.device)
               
                # tetrahedral FEM
                if (model.tet_count):

                    wp.launch(kernel=solve_tetrahedra,
                                dim=model.tet_count,
                                inputs=[q_pred, qd_pred, model.particle_inv_mass, model.tet_indices, model.tet_poses, model.tet_activations, model.tet_materials, dt, self.relaxation],
                                outputs=[state_out.particle_f],
                                device=model.device)

                # apply updates
                wp.launch(kernel=apply_deltas,
                            dim=model.particle_count,
                            inputs=[state_in.particle_q,
                                    state_in.particle_qd,
                                    q_pred,
                                    state_out.particle_f,
                                    dt],
                            outputs=[q_pred,
                                     qd_pred],
                            device=model.device)

            state_out.particle_q = q_pred
            state_out.particle_qd = qd_pred

            return state_out




# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import warp as wp
from warp.tests.test_base import *

import numpy as np

compare_to_numpy = False
print_results = False

wp.init()

@wp.kernel
def test_kernel(
    x: wp.array(dtype=float),
    x_round: wp.array(dtype=float),
    x_rint: wp.array(dtype=float),
    x_trunc: wp.array(dtype=float),
    x_cast: wp.array(dtype=float),
    x_floor: wp.array(dtype=float),
    x_ceil: wp.array(dtype=float)):

    tid = wp.tid()

    x_round[tid] = wp.round(x[tid])
    x_rint[tid] = wp.rint(x[tid])
    x_trunc[tid] = wp.trunc(x[tid])
    x_cast[tid] = float(int(x[tid]))
    x_floor[tid] = wp.floor(x[tid])
    x_ceil[tid] = wp.ceil(x[tid])

def test_rounding(test, device):
        
    nx = np.array([ 4.9,  4.5,  4.1,  3.9,  3.5,  3.1,  2.9,  2.5,  2.1,  1.9,  1.5,  1.1,  0.9,  0.5,  0.1,
                -0.1, -0.5, -0.9, -1.1, -1.5, -1.9, -2.1, -2.5, -2.9, -3.1, -3.5, -3.9, -4.1, -4.5, -4.9],
                dtype=np.float32)

    x = wp.array(nx, device=device)
    N = len(x)

    x_round = wp.empty(N, dtype=float, device=device)
    x_rint = wp.empty(N, dtype=float, device=device)
    x_trunc = wp.empty(N, dtype=float, device=device)
    x_cast = wp.empty(N, dtype=float, device=device)
    x_floor = wp.empty(N, dtype=float, device=device)
    x_ceil = wp.empty(N, dtype=float, device=device)

    wp.launch(
        kernel=test_kernel,
        dim=N,
        inputs=[x, x_round, x_rint, x_trunc, x_cast, x_floor, x_ceil],
        device=device
    )

    wp.synchronize()

    nx_round = x_round.numpy().reshape(N)
    nx_rint = x_rint.numpy().reshape(N)
    nx_trunc = x_trunc.numpy().reshape(N)
    nx_cast = x_cast.numpy().reshape(N)
    nx_floor = x_floor.numpy().reshape(N)
    nx_ceil = x_ceil.numpy().reshape(N)

    tab = np.stack([nx, nx_round, nx_rint, nx_trunc, nx_cast, nx_floor, nx_ceil], axis=1)

    golden = np.array(
       [[  4.9,  5.,  5.,  4.,  4.,  4.,  5.],
        [  4.5,  5.,  4.,  4.,  4.,  4.,  5.],
        [  4.1,  4.,  4.,  4.,  4.,  4.,  5.],
        [  3.9,  4.,  4.,  3.,  3.,  3.,  4.],
        [  3.5,  4.,  4.,  3.,  3.,  3.,  4.],
        [  3.1,  3.,  3.,  3.,  3.,  3.,  4.],
        [  2.9,  3.,  3.,  2.,  2.,  2.,  3.],
        [  2.5,  3.,  2.,  2.,  2.,  2.,  3.],
        [  2.1,  2.,  2.,  2.,  2.,  2.,  3.],
        [  1.9,  2.,  2.,  1.,  1.,  1.,  2.],
        [  1.5,  2.,  2.,  1.,  1.,  1.,  2.],
        [  1.1,  1.,  1.,  1.,  1.,  1.,  2.],
        [  0.9,  1.,  1.,  0.,  0.,  0.,  1.],
        [  0.5,  1.,  0.,  0.,  0.,  0.,  1.],
        [  0.1,  0.,  0.,  0.,  0.,  0.,  1.],
        [ -0.1, -0., -0., -0.,  0., -1., -0.],
        [ -0.5, -1., -0., -0.,  0., -1., -0.],
        [ -0.9, -1., -1., -0.,  0., -1., -0.],
        [ -1.1, -1., -1., -1., -1., -2., -1.],
        [ -1.5, -2., -2., -1., -1., -2., -1.],
        [ -1.9, -2., -2., -1., -1., -2., -1.],
        [ -2.1, -2., -2., -2., -2., -3., -2.],
        [ -2.5, -3., -2., -2., -2., -3., -2.],
        [ -2.9, -3., -3., -2., -2., -3., -2.],
        [ -3.1, -3., -3., -3., -3., -4., -3.],
        [ -3.5, -4., -4., -3., -3., -4., -3.],
        [ -3.9, -4., -4., -3., -3., -4., -3.],
        [ -4.1, -4., -4., -4., -4., -5., -4.],
        [ -4.5, -5., -4., -4., -4., -5., -4.],
        [ -4.9, -5., -5., -4., -4., -5., -4.]], dtype=np.float32)

    assert_np_equal(tab, golden)

    if (print_results):
        np.set_printoptions(formatter={'float': lambda x: "{:6.1f}".format(x).replace(".0", ".")})

        print("----------------------------------------------")
        print("   %5s %5s %5s %5s %5s %5s %5s" % ("x ", "round", "rint", "trunc", "cast", "floor", "ceil"))
        print(tab)
        print("----------------------------------------------")

    if compare_to_numpy:
        nx_round = np.round(nx)
        nx_rint = np.rint(nx)
        nx_trunc = np.trunc(nx)
        nx_fix = np.fix(nx)
        nx_floor = np.floor(nx)
        nx_ceil = np.ceil(nx)

        tab = np.stack([nx, nx_round, nx_rint, nx_trunc, nx_fix, nx_floor, nx_ceil], axis=1)
        print("   %5s %5s %5s %5s %5s %5s %5s" % ("x ", "round", "rint", "trunc", "fix", "floor", "ceil"))
        print(tab)
        print("----------------------------------------------")


def register(parent):

    class TestRounding(parent):
        pass

    add_function_test(TestRounding, "test_rounding", test_rounding, devices=wp.get_devices())

    return TestRounding

if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2)


# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

# include parent path
import numpy as np
import math

import warp as wp
from warp.tests.test_base import *

np.random.seed(42)

wp.init()

@wp.kernel
def sample_mesh_query(mesh: wp.uint64,
                query_points: wp.array(dtype=wp.vec3),
                query_faces: wp.array(dtype=int),
                query_signs: wp.array(dtype=float),
                query_dist: wp.array(dtype=float)):
    
    tid = wp.tid()

    face_index = int(0)
    face_u = float(0.0)
    face_v = float(0.0)
    sign = float(0.0)

    max_dist = 10012.0

    p = query_points[tid]
    
    wp.mesh_query_point(mesh, p, max_dist, sign, face_index, face_u, face_v)
        
    cp = wp.mesh_eval_position(mesh, face_index, face_u, face_v)

    query_signs[tid] = sign
    query_faces[tid] = face_index
    query_dist[tid] = wp.length(cp-p)


@wp.func
def triangle_closest_point(a: wp.vec3, b: wp.vec3, c: wp.vec3, p: wp.vec3):
    ab = b - a
    ac = c - a
    ap = p - a

    d1 = wp.dot(ab, ap)
    d2 = wp.dot(ac, ap)

    if (d1 <= 0.0 and d2 <= 0.0):
        return wp.vec2(1.0, 0.0)

    bp = p - b
    d3 = wp.dot(ab, bp)
    d4 = wp.dot(ac, bp)

    if (d3 >= 0.0 and d4 <= d3):
        return wp.vec2(0.0, 1.0)

    vc = d1 * d4 - d3 * d2
    v = d1 / (d1 - d3)
    if (vc <= 0.0 and d1 >= 0.0 and d3 <= 0.0):
        return wp.vec2(1.0 - v, v)

    cp = p - c
    d5 = wp.dot(ab, cp)
    d6 = wp.dot(ac, cp)

    if (d6 >= 0.0 and d5 <= d6):
        return wp.vec2(0.0, 0.0)

    vb = d5 * d2 - d1 * d6
    w = d2 / (d2 - d6)
    if (vb <= 0.0 and d2 >= 0.0 and d6 <= 0.0):
        return wp.vec2(1.0 - w, 0.0)

    va = d3 * d6 - d5 * d4
    w = (d4 - d3) / ((d4 - d3) + (d5 - d6))
    if (va <= 0.0 and (d4 - d3) >= 0.0 and (d5 - d6) >= 0.0):
        return wp.vec2(0.0, 1.0 - w)

    denom = 1.0 / (va + vb + vc)
    v = vb * denom
    w = vc * denom
    u = 1.0 - v - w

    return wp.vec2(u, v)


@wp.func
def solid_angle(v0: wp.vec3, v1: wp.vec3, v2: wp.vec3, p: wp.vec3):
    a = v0 - p
    b = v1 - p
    c = v2 - p

    a_len = wp.length(a)
    b_len = wp.length(b)
    c_len = wp.length(c)

    det = wp.dot(a, wp.cross(b, c))
    den = a_len*b_len*c_len + wp.dot(a, b)*c_len + wp.dot(b, c)*a_len + wp.dot(c, a)*b_len

    return wp.atan2(det, den)



@wp.kernel
def sample_mesh_brute(
                tri_points: wp.array(dtype=wp.vec3),
                tri_indices: wp.array(dtype=int),
                tri_count: int,
                query_points: wp.array(dtype=wp.vec3),
                query_faces: wp.array(dtype=int),
                query_signs: wp.array(dtype=float),
                query_dist: wp.array(dtype=float)):
    
    tid = wp.tid()

    min_face = int(0)
    min_dist = float(1.e+6)

    winding_angle = float(0.0)

    p = query_points[tid]

    for i in range(0, tri_count):
        
        a = tri_points[tri_indices[i*3+0]]
        b = tri_points[tri_indices[i*3+1]]
        c = tri_points[tri_indices[i*3+2]]

        winding_angle += solid_angle(a, b, c, p)

        bary = triangle_closest_point(a, b, c, p)
        u = bary[0]
        v = bary[1]

        cp = u*a + v*b + (1.0-u-v)*c
        cp_dist = wp.length(cp - p)

        if (cp_dist < min_dist):
            min_dist = cp_dist
            min_face = i
            
    query_faces[tid] = min_face
    query_signs[tid] = winding_angle 
    query_dist[tid] = min_dist


# constructs a grid of evenly spaced particles
def particle_grid(dim_x, dim_y, dim_z, lower, radius, jitter):
    points = np.meshgrid(np.linspace(0, dim_x, dim_x), np.linspace(0, dim_y, dim_y), np.linspace(0, dim_z, dim_z))
    points_t = np.array((points[0], points[1], points[2])).T*radius*2.0 + np.array(lower)
    points_t = points_t + np.random.rand(*points_t.shape)*radius*jitter

    return points_t.reshape((-1, 3))

# triangulate a list of polygon face indices
def triangulate(face_counts, face_indices):
    num_tris = np.sum(np.subtract(face_counts, 2))
    num_tri_vtx = num_tris * 3
    tri_indices = np.zeros(num_tri_vtx, dtype=int)
    ctr = 0
    wedgeIdx = 0

    for nb in face_counts:
        for i in range(nb-2):
            tri_indices[ctr] = face_indices[wedgeIdx]
            tri_indices[ctr + 1] = face_indices[wedgeIdx + i + 1]
            tri_indices[ctr + 2] = face_indices[wedgeIdx + i + 2]
            ctr+=3
        wedgeIdx+=nb

    return tri_indices


def test_mesh_query_point(test, device):

    from pxr import Usd, UsdGeom

    mesh = Usd.Stage.Open(os.path.abspath(os.path.join(os.path.dirname(__file__), "assets/spiky.usd")))
    mesh_geom = UsdGeom.Mesh(mesh.GetPrimAtPath("/Cube/Cube"))

    mesh_counts = mesh_geom.GetFaceVertexCountsAttr().Get()
    mesh_indices = mesh_geom.GetFaceVertexIndicesAttr().Get()

    tri_indices = triangulate(mesh_counts, mesh_indices)

    mesh_points = wp.array(np.array(mesh_geom.GetPointsAttr().Get()), dtype=wp.vec3, device=device)
    mesh_indices = wp.array(np.array(tri_indices), dtype=int, device=device)

    # create mesh
    mesh = wp.Mesh(
        points=mesh_points, 
        velocities=None,
        indices=mesh_indices)

    p = particle_grid(32, 32, 32, np.array([-1.1, -1.1, -1.1]), 0.05, 0.0)

    query_count = len(p)
    query_points = wp.array(p, dtype=wp.vec3, device=device)

    signs_query = wp.zeros(query_count, dtype=float, device=device)
    faces_query = wp.zeros(query_count, dtype=int, device=device)
    dist_query = wp.zeros(query_count, dtype=float, device=device)

    signs_brute = wp.zeros(query_count, dtype=float, device=device)
    faces_brute = wp.zeros(query_count, dtype=int, device=device)
    dist_brute = wp.zeros(query_count, dtype=float, device=device)

    wp.launch(kernel=sample_mesh_query, dim=query_count, inputs=[mesh.id, query_points, faces_query, signs_query, dist_query], device=device)
    wp.launch(kernel=sample_mesh_brute, dim=query_count, inputs=[mesh_points, mesh_indices, int(len(mesh_indices)/3), query_points, faces_brute, signs_brute, dist_brute], device=device)

    signs_query = signs_query.numpy()
    faces_query = faces_query.numpy()
    dist_query = dist_query.numpy()

    signs_brute = signs_brute.numpy()
    faces_brute = faces_brute.numpy()
    dist_brute = dist_brute.numpy()

    query_points = query_points.numpy()

    inside_query = []
    inside_brute = []

    for i in range(query_count):

        if (signs_query[i] < 0.0):
            inside_query.append(query_points[i].tolist())
        
        if (signs_brute[i] > 6.0):
            inside_brute.append(query_points[i].tolist())

    inside_query = np.array(inside_query)
    inside_brute = np.array(inside_brute)

    # import warp.render

    # stage = warp.render.UsdRenderer("tests/outputs/test_mesh_query_point.usd")

    # radius = 0.1
    # stage.begin_frame(0.0)
    # stage.render_mesh(points=mesh_points.numpy(), indices=mesh_indices.numpy(), name="mesh")
    # stage.render_points(points=inside_query, radius=radius, name="query")
    # stage.render_points(points=inside_brute, radius=radius, name="brute")
    # stage.render_points(points=query_points, radius=radius, name="all")
    # stage.end_frame()

    # stage.save()

    test.assertTrue(len(inside_query) == len(inside_brute))

    dist_error = np.max(np.abs(dist_query - dist_brute))
    sign_error = np.max(np.abs(inside_query - inside_brute))

    tolerance = 1.5e-4
    test.assertTrue(dist_error < tolerance, f"dist_error is {dist_error} which is >= {tolerance}")
    test.assertTrue(sign_error < tolerance, f"sign_error is {sign_error} which is >= {tolerance}")



@wp.kernel
def mesh_query_point_loss(mesh: wp.uint64,
                query_points: wp.array(dtype=wp.vec3),
                projected_points: wp.array(dtype=wp.vec3),
                loss: wp.array(dtype=float)):
    
    tid = wp.tid()

    face_index = int(0)
    face_u = float(0.0)
    face_v = float(0.0)
    sign = float(0.0)

    max_dist = 10012.0


    p = query_points[tid]
    
    wp.mesh_query_point(mesh, p, max_dist, sign, face_index, face_u, face_v)
    q = wp.mesh_eval_position(mesh, face_index, face_u, face_v)

    projected_points[tid] = q

    dist = wp.length(wp.sub(p, q))
    loss[tid] = dist


def test_adj_mesh_query_point(test, device):

    from pxr import Usd, UsdGeom

    mesh = Usd.Stage.Open(os.path.abspath(os.path.join(os.path.dirname(__file__), "assets/torus.usda")))
    mesh_geom = UsdGeom.Mesh(mesh.GetPrimAtPath("/World/Torus"))

    mesh_counts = mesh_geom.GetFaceVertexCountsAttr().Get()
    mesh_indices = mesh_geom.GetFaceVertexIndicesAttr().Get()

    tri_indices = triangulate(mesh_counts, mesh_indices)

    mesh_points = wp.array(np.array(mesh_geom.GetPointsAttr().Get()), dtype=wp.vec3, device=device)
    mesh_indices = wp.array(np.array(tri_indices), dtype=int, device=device)

    # test tri
    # print("Testing Single Triangle")
    # mesh_points = wp.array(np.array([[0.0, 0.0, 0.0], [2.0, 0.0, 0.0], [0.0, 2.0, 0.0]]), dtype=wp.vec3, device=device)
    # mesh_indices = wp.array(np.array([0,1,2]), dtype=int, device=device)

    # create mesh
    mesh = wp.Mesh(
        points=mesh_points, 
        velocities=None,
        indices=mesh_indices)

    # p = particle_grid(32, 32, 32, np.array([-5.0, -5.0, -5.0]), 0.1, 0.1)*100.0
    p = wp.vec3(50.0, 50.0, 50.0)

    tape = wp.Tape()

    # analytic gradients
    with tape:

        query_points = wp.array(p, dtype=wp.vec3, device=device, requires_grad=True)
        projected_points = wp.zeros(n=1, dtype=wp.vec3, device=device)
        loss = wp.zeros(n=1, dtype=float, device=device, requires_grad=True)

        wp.launch(kernel=mesh_query_point_loss, dim=1, inputs=[mesh.id, query_points, projected_points, loss], device=device)

    tape.backward(loss=loss)
    analytic = tape.gradients[query_points].numpy().flatten()

    # numeric gradients
    eps = 1.e-3
    loss_values = []
    numeric = np.zeros(3)

    offset_query_points = [
        wp.vec3(p[0] - eps, p[1], p[2]), wp.vec3(p[0] + eps, p[1], p[2]),
        wp.vec3(p[0], p[1] - eps, p[2]), wp.vec3(p[0], p[1] + eps, p[2]),
        wp.vec3(p[0], p[1], p[2] - eps), wp.vec3(p[0], p[1], p[2] + eps)]

    for i in range(6):
        q = offset_query_points[i]

        query_points = wp.array(q, dtype=wp.vec3, device=device)
        projected_points = wp.zeros(n=1, dtype=wp.vec3, device=device)
        loss = wp.zeros(n=1, dtype=float, device=device)

        wp.launch(kernel=mesh_query_point_loss, dim=1, inputs=[mesh.id, query_points, projected_points, loss], device=device)

        loss_values.append(loss.numpy()[0])

    for i in range(3):
        l_0 = loss_values[i*2]
        l_1 = loss_values[i*2+1]
        gradient = (l_1 - l_0) / (2.0*eps)
        numeric[i] = gradient

    error = ((analytic - numeric) * (analytic - numeric)).sum(axis=0)

    tolerance = 1.e-3
    test.assertTrue(error < tolerance, f"error is {error} which is >= {tolerance}")


def register(parent):

    devices = wp.get_devices()

    class TestMeshQuery(parent):
        pass

    # USD import failures should not count as a test failure
    try:
        from pxr import Usd, UsdGeom
        have_usd = True
    except:
        have_usd = False

    if have_usd:
        add_function_test(TestMeshQuery, "test_mesh_query_point", test_mesh_query_point, devices=devices)
        add_function_test(TestMeshQuery, "test_adj_mesh_query_point", test_adj_mesh_query_point, devices=devices)

    return TestMeshQuery

if __name__ == '__main__':
    c = register(unittest.TestCase)

    unittest.main(verbosity=2)



# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

# include parent path
import numpy as np
import math

import warp as wp
from warp.tests.test_base import *

import unittest

wp.init()

@wp.kernel
def add_vec2(dest: wp.array(dtype=wp.vec2),
             c: wp.vec2):

    tid = wp.tid()
    dest[tid] = c


@wp.kernel
def transform_vec2(dest: wp.array(dtype=wp.vec2),
    m: wp.mat22,
    v: wp.vec2):

    tid = wp.tid()

    p = wp.mul(m, v)
    dest[tid] = p


@wp.kernel
def add_vec3(dest: wp.array(dtype=wp.vec3),
             c: wp.vec3):

    tid = wp.tid()
    dest[tid] = c

@wp.kernel
def transform_vec3(dest: wp.array(dtype=wp.vec3),
    m: wp.mat33,
    v: wp.vec3):

    tid = wp.tid()

    p = wp.mul(m, v)
    dest[tid] = p


@wp.kernel
def transform_multiply(xforms: wp.array(dtype=wp.transform),
    a: wp.transform):

    tid = wp.tid()

    xforms[tid] = wp.transform_multiply(xforms[tid], a)


def test_vec2_arg(test, device, n):
    
    dest = wp.zeros(n=n, dtype=wp.vec2, device=device)
    c = np.array((1.0, 2.0))

    wp.launch(add_vec2, dim=n, inputs=[dest, c], device=device)

    # ensure type can round-trip from Python->GPU->Python
    test.assertTrue(np.array_equal(dest.numpy(), np.tile(c, (n, 1))))

def test_vec2_transform(test, device, n):

    dest = wp.zeros(n=n, dtype=wp.vec2, device=device)
    c = np.array((1.0, 2.0))
    m = np.array(((3.0, -1.0),
                  (2.5, 4.0)))

    wp.launch(transform_vec2, dim=n, inputs=[dest, m, c], device=device)
    test.assertTrue(np.array_equal(dest.numpy(), np.tile(m@c, (n, 1))))

def test_vec3_arg(test, device, n):
        
    dest = wp.zeros(n=n, dtype=wp.vec3, device=device)
    c = np.array((1.0, 2.0, 3.0))

    wp.launch(add_vec3, dim=n, inputs=[dest, c], device=device)
    test.assertTrue(np.array_equal(dest.numpy(), np.tile(c, (n, 1))))

def test_vec3_transform(test, device, n):

    dest = wp.zeros(n=n, dtype=wp.vec3, device=device)
    c = np.array((1.0, 2.0, 3.0))
    m = np.array(((1.0, 2.0, 3.0),
                  (4.0, 5.0, 6.0),
                  (7.0, 8.0, 9.0)))

    wp.launch(transform_vec3, dim=n, inputs=[dest, m, c], device=device)
    test.assertTrue(np.array_equal(dest.numpy(), np.tile(m@c, (n, 1))))

def test_transform_multiply(test, device, n):

    a = wp.transform((0.0, 1.0, 0.0), wp.utils.quat_identity())

    x = []
    for i in range(10):
        x.append(wp.utils.transform_identity())

    xforms = wp.array(x, dtype=wp.transform, device=device)
    wp.launch(transform_multiply, dim=n, inputs=[xforms, a], device=device)


# construct kernel + test harness for given matrix / vector types
def make_matrix_test(dim, matrix, vector):
   
    def test_matrix_kernel(a: wp.array(dtype=matrix),
                           b: wp.array(dtype=matrix),
                           c: wp.array(dtype=matrix),               
                           x: wp.array(dtype=vector),
                           result_m: wp.array(dtype=matrix),
                           result_i: wp.array(dtype=matrix),
                           result_d: wp.array(dtype=float),
                           result_x: wp.array(dtype=vector)):

        tid = wp.tid()

        m = a[tid]*b[tid] + c[tid]*2.0
        
        result_m[tid] = m
        result_x[tid] = m*x[tid]

        result_d[tid] = wp.determinant(m)

        invm = wp.inverse(m)
        result_i[tid] = m*invm


    # register a custom kernel (no decorator) function
    # this lets us register the same function definition
    # against multiple symbols, with different arg types
    module = wp.get_module(test_matrix_kernel.__module__)
    kernel = wp.Kernel(func=test_matrix_kernel, key=f"test_mat{dim}{dim}_kernel", module=module)
        
    def test_matrix(test, device):

        rng = np.random.default_rng(42)

        n = 1024

        a = rng.random(size=(n, dim, dim), dtype=float)
        b = rng.random(size=(n, dim, dim), dtype=float)
        c = rng.random(size=(n, dim, dim), dtype=float)
        x = rng.random(size=(n, dim, 1), dtype=float)

        a_array = wp.array(a, dtype=matrix, device=device)
        b_array = wp.array(b, dtype=matrix, device=device)
        c_array = wp.array(c, dtype=matrix, device=device)

        x_array = wp.array(x, dtype=vector, device=device)

        result_m_array = wp.zeros_like(a_array)
        result_i_array = wp.zeros_like(a_array)
        result_x_array = wp.zeros_like(x_array)
        result_d_array = wp.zeros(n, dtype=float, device=device)
        
        wp.launch(kernel, n, inputs=[a_array, b_array, c_array, x_array, result_m_array, result_i_array, result_d_array, result_x_array], device=device)

        # numpy reference result
        result_m = np.matmul(a,b) + c*2.0
        result_x = np.matmul(result_m, x)
        result_i = np.array([np.eye(dim)]*n)
        result_d = np.linalg.det(result_m)

        assert_np_equal(result_m_array.numpy(), result_m, tol=1.e-5)
        assert_np_equal(result_i_array.numpy(), result_i, tol=1.e-3)
        assert_np_equal(result_d_array.numpy(), result_d, tol=1.e-3)
        assert_np_equal(result_x_array.numpy(), result_x, tol=1.e-5)


    return test_matrix


# generate test functions for matrix types
test_mat22 = make_matrix_test(2, wp.mat22, wp.vec2)
test_mat33 = make_matrix_test(3, wp.mat33, wp.vec3)
test_mat44 = make_matrix_test(4, wp.mat44, wp.vec4)


def test_scalar_array(test, device):

    scalar_list = (0.0, 1.0, 2.0)
    scalar_array = wp.array(scalar_list, device=device)

    assert_np_equal(np.array(scalar_list), scalar_array.numpy())

def test_vector_array(test, device):

    vector_list = [ (0.0, 0.0, 0.0),
                    (1.0, 1.0, 1.0),
                    (2.0, 2.0, 2.0) ]

    vector_array = wp.array(vector_list, dtype=wp.vec3, device=device)

    assert_np_equal(np.array(vector_list), vector_array.numpy())
    


@wp.kernel
def test_vector_arg_types(v2: wp.vec2,
                          v3: wp.vec3,
                          v4: wp.vec4,
                          m22: wp.mat22,
                          m33: wp.mat33,
                          m44: wp.mat44):

    wp.expect_eq(v2, wp.vec2(1.0, 2.0))
    wp.expect_eq(v3, wp.vec3(1.0, 2.0, 3.0))
    wp.expect_eq(v4, wp.vec4(1.0, 2.0, 3.0, 4.0))
    
    wp.expect_eq(m22, wp.mat22(1.0, 2.0, 
                               3.0, 4.0))
    
    wp.expect_eq(m33, wp.mat33(1.0, 2.0, 3.0,
                               4.0, 5.0, 6.0,
                               7.0, 8.0, 9.0))

    wp.expect_eq(m44, wp.mat44(1.0, 2.0, 3.0, 4.0,
                               5.0, 6.0, 7.0, 8.0,
                               9.0, 10.0, 11.0, 12.0,
                               13.0, 14.0, 15.0, 16.0))


@wp.kernel
def test_scalar_arg_types(i8: wp.int8,
                          u8: wp.uint8,
                          i16: wp.int16,
                          u16: wp.uint16,
                          i32: wp.int32,                  
                          u32: wp.uint32,
                          i64: wp.int64,
                          u64: wp.uint64,
                          f32: wp.float32,
                          f64: wp.float64):
                  
    wp.expect_eq(int(i8), -64)
    wp.expect_eq(int(u8),  255)
    wp.expect_eq(int(i16), -64)
    wp.expect_eq(int(u16),  255)
    wp.expect_eq(int(i32), -64)
    wp.expect_eq(int(u32), 255)
    wp.expect_eq(int(i64), -64)
    wp.expect_eq(int(u64), 255)
    wp.expect_eq(int(f32), 3)
    wp.expect_eq(int(f64), 3)
    wp.expect_eq(float(f32), 3.14159)
    wp.expect_eq(float(f64), 3.14159)


@wp.kernel
def test_scalar_array_types_load(i8: wp.array(dtype=wp.int8),
                                 u8: wp.array(dtype=wp.uint8),
                                 i16: wp.array(dtype=wp.int16),
                                 u16: wp.array(dtype=wp.uint16),
                                 i32: wp.array(dtype=wp.int32),
                                 u32: wp.array(dtype=wp.uint32),
                                 i64: wp.array(dtype=wp.int64),
                                 u64: wp.array(dtype=wp.uint64),
                                 f32: wp.array(dtype=wp.float32),
                                 f64: wp.array(dtype=wp.float64)):
                  
    tid = wp.tid()

    wp.expect_eq(int(i8[tid]), tid)
    wp.expect_eq(int(u8[tid]), tid)
    wp.expect_eq(int(i16[tid]), tid)
    wp.expect_eq(int(u16[tid]), tid)
    wp.expect_eq(int(i32[tid]), tid)
    wp.expect_eq(int(u32[tid]), tid)
    wp.expect_eq(int(i64[tid]), tid)
    wp.expect_eq(int(u64[tid]), tid)
    wp.expect_eq(float(f32[tid]), float(tid))
    wp.expect_eq(float(f64[tid]), float(tid))
    
@wp.kernel
def test_scalar_array_types_store(i8: wp.array(dtype=wp.int8),
                                  u8: wp.array(dtype=wp.uint8),
                                  i16: wp.array(dtype=wp.int16),
                                  u16: wp.array(dtype=wp.uint16),                                  
                                  i32: wp.array(dtype=wp.int32),
                                  u32: wp.array(dtype=wp.uint32),
                                  i64: wp.array(dtype=wp.int64),
                                  u64: wp.array(dtype=wp.uint64),
                                  f32: wp.array(dtype=wp.float32),
                                  f64: wp.array(dtype=wp.float64)):
                  
    tid = wp.tid()

    i8[tid] = wp.int8(tid)
    u8[tid] = wp.uint8(tid)
    i16[tid] = wp.int16(tid)
    u16[tid] = wp.uint16(tid)
    i32[tid] = wp.int32(tid)
    u32[tid] = wp.uint32(tid)
    i64[tid] = wp.int64(tid)
    u64[tid] = wp.uint64(tid)
    f32[tid] = wp.float32(tid)
    f64[tid] = wp.float64(tid)

    # check round-trip
    wp.expect_eq(int(i8[tid]), tid)
    wp.expect_eq(int(u8[tid]), tid)
    wp.expect_eq(int(i16[tid]), tid)
    wp.expect_eq(int(u16[tid]), tid)
    wp.expect_eq(int(i32[tid]), tid)
    wp.expect_eq(int(u32[tid]), tid)
    wp.expect_eq(int(i64[tid]), tid)
    wp.expect_eq(int(u64[tid]), tid)
    wp.expect_eq(float(f32[tid]), float(tid))
    wp.expect_eq(float(f64[tid]), float(tid))


def test_scalar_array_types(test, device, load, store):

    dim = 64

    i8 = wp.array(np.linspace(0, dim, dim,  endpoint=False, dtype=np.int8), device=device)
    u8 = wp.array(np.linspace(0, dim, dim,  endpoint=False, dtype=np.uint8), device=device)
    i16 = wp.array(np.linspace(0, dim, dim, endpoint=False, dtype=np.int16), device=device)
    u16 = wp.array(np.linspace(0, dim, dim, endpoint=False, dtype=np.uint16), device=device)
    i32 = wp.array(np.linspace(0, dim, dim, endpoint=False, dtype=np.int32), device=device)
    u32 = wp.array(np.linspace(0, dim, dim, endpoint=False, dtype=np.uint32), device=device)
    i64 = wp.array(np.linspace(0, dim, dim, endpoint=False, dtype=np.int64), device=device)
    u64 = wp.array(np.linspace(0, dim, dim, endpoint=False, dtype=np.uint64), device=device)
    f32 = wp.array(np.linspace(0, dim, dim, endpoint=False, dtype=np.float32), device=device)
    f64 = wp.array(np.linspace(0, dim, dim, endpoint=False, dtype=np.float64), device=device)

    if load:
        wp.launch(test_scalar_array_types_load, dim=dim, inputs=[i8, u8, i16, u16, i32, u32, i64, u64, f32, f64], device=device)

    if store:
        wp.launch(test_scalar_array_types_store, dim=dim, inputs=[i8, u8, i16, u16, i32, u32, i64, u64, f32, f64], device=device)


@wp.kernel
def test_transform_matrix():

    r = wp.quat_from_axis_angle(wp.vec3(1.0, 0.0, 0.0), 0.5)
    t = wp.vec3(0.25, 0.5, -0.75)
    s = wp.vec3(2.0, 0.5, 0.75)

    m = wp.mat44(t, r, s)

    p = wp.vec3(1.0, 2.0, 3.0)

    r_0 = wp.quat_rotate(r, wp.cw_mul(s, p)) + t
    r_1 = wp.transform_point(m, p)

    r_2 = wp.transform_vector(m, p)

    wp.expect_near(r_0, r_1, 1.e-4)
    wp.expect_near(r_2, r_0 - t, 1.e-4)


def register(parent):

    devices = wp.get_devices()

    class TestCTypes(parent):
        pass

    inputs = [wp.vec2(1.0, 2.0),
            wp.vec3(1.0, 2.0, 3.0),
            wp.vec4(1.0, 2.0, 3.0, 4.0),
            wp.mat22(1.0, 2.0,
                    3.0, 4.0),
            wp.mat33(1.0, 2.0, 3.0,
                    4.0, 5.0, 6.0,
                    7.0, 8.0, 9.0),
            wp.mat44(1.0, 2.0, 3.0, 4.0,
                    5.0, 6.0, 7.0, 8.0,
                    9.0, 10.0, 11.0, 12.0,
                    13.0, 14.0, 15.0, 16.0)]


    add_function_test(TestCTypes, "test_mat22", test_mat22, devices=devices)
    add_function_test(TestCTypes, "test_mat33", test_mat33, devices=devices)
    add_function_test(TestCTypes, "test_mat44", test_mat44, devices=devices)
    add_kernel_test(TestCTypes, name="test_scalar_arg_types", kernel=test_scalar_arg_types, dim=1, inputs=[-64, 255, -64, 255, -64, 255, -64, 255, 3.14159, 3.14159], devices=devices)
    add_kernel_test(TestCTypes, name="test_vector_arg_types", kernel=test_vector_arg_types, dim=1, inputs=inputs, devices=devices)
    add_function_test(TestCTypes, "test_scalar_array_load", test_scalar_array_types, devices=devices, load=True, store=False)
    add_function_test(TestCTypes, "test_scalar_array_store", test_scalar_array_types, devices=devices, load=False, store=True)
    add_function_test(TestCTypes, "test_vec2_arg", test_vec2_arg, devices=devices, n=8)
    add_function_test(TestCTypes, "test_vec2_transform", test_vec2_transform, devices=devices, n=8)
    add_function_test(TestCTypes, "test_vec3_arg", test_vec3_arg, devices=devices, n=8)
    add_function_test(TestCTypes, "test_vec3_transform", test_vec3_transform, devices=devices, n=8)
    add_function_test(TestCTypes, "test_transform_multiply", test_transform_multiply, devices=devices, n=8)
    add_kernel_test(TestCTypes, name="test_transform_matrix", kernel=test_transform_matrix, dim=1, devices=devices)
    add_function_test(TestCTypes, "test_scalar_array", test_scalar_array, devices=devices)
    add_function_test(TestCTypes, "test_vector_array", test_vector_array, devices=devices)

    return TestCTypes

if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2)


import setuptools

setuptools.setup(
    name="warp-lang",
    version="0.5.0",
    author="NVIDIA",
    author_email="mmacklin@nvidia.com",
    description="A Python framework for high-performance simulation and graphics programming",
    url="https://github.com/NVIDIA/warp",
    project_urls={
        "Documentation": "https://nvidia.github.io/warp",
    },
    long_description="",
    long_description_content_type="text/markdown",
    license="NVSCL",
    packages=setuptools.find_packages(),
    package_data={"": ["native/*", "native/nanovdb/*", "tests/assets/*", "bin/*", "gen/*"]},
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
    ],
    install_requires=["numpy"],
    python_requires=">=3.7"
)


# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

# include parent path
import numpy as np
import math

import warp as wp
from warp.tests.test_base import *

import unittest

wp.init()

# construct kernel + test function for atomic ops on each vec/matrix type
def make_atomic_test(type):
   
    def test_atomic_kernel(out_add: wp.array(dtype=type),
                           out_min: wp.array(dtype=type),
                           out_max: wp.array(dtype=type),
                           val: wp.array(dtype=type)):

        tid = wp.tid()
        
        wp.atomic_add(out_add, 0, val[tid])
        wp.atomic_min(out_min, 0, val[tid])
        wp.atomic_max(out_max, 0, val[tid])


    # register a custom kernel (no decorator) function
    # this lets us register the same function definition
    # against multiple symbols, with different arg types
    module = wp.get_module(test_atomic_kernel.__module__)
    kernel = wp.Kernel(func=test_atomic_kernel, key=f"test_atomic_{type.__name__}_kernel", module=module)
        
    def test_atomic(test, device):

        n = 1024

        rng = np.random.default_rng(42)
        
        if type == wp.int32:
            base = (rng.random(size=1, dtype=np.float32)*100.0).astype(np.int32)
            val = (rng.random(size=n, dtype=np.float32)*100.0).astype(np.int32)
        
        elif type == wp.float32:
            base = rng.random(size=1, dtype=np.float32)
            val = rng.random(size=n, dtype=np.float32)
        
        else:
            base = rng.random(size=(1, *type._shape_), dtype=float)
            val = rng.random(size=(n, *type._shape_), dtype=float)
        
        add_array = wp.array(base, dtype=type, device=device)
        min_array = wp.array(base, dtype=type, device=device)
        max_array = wp.array(base, dtype=type, device=device)
        
        val_array = wp.array(val, dtype=type, device=device)

        wp.launch(kernel, n, inputs=[add_array, min_array, max_array, val_array], device=device)

        val = np.append(val, [base[0]], axis=0)

        assert_np_equal(add_array.numpy(), np.sum(val, axis=0), tol=1.e-2)
        assert_np_equal(min_array.numpy(), np.min(val, axis=0), tol=1.e-2)
        assert_np_equal(max_array.numpy(), np.max(val, axis=0), tol=1.e-2)

    return test_atomic

# generate test functions for atomic types
test_atomic_int = make_atomic_test(wp.int32)
test_atomic_float = make_atomic_test(wp.float32)
test_atomic_vec2 = make_atomic_test(wp.vec2)
test_atomic_vec3 = make_atomic_test(wp.vec3)
test_atomic_vec4 = make_atomic_test(wp.vec4)
test_atomic_mat22 = make_atomic_test(wp.mat22)
test_atomic_mat33 = make_atomic_test(wp.mat33)
test_atomic_mat44 = make_atomic_test(wp.mat44)


def register(parent):

    devices = wp.get_devices()

    class TestAtomic(parent):
        pass
    
    add_function_test(TestAtomic, "test_atomic_int", test_atomic_int, devices=devices)
    add_function_test(TestAtomic, "test_atomic_float", test_atomic_float, devices=devices)
    add_function_test(TestAtomic, "test_atomic_vec2", test_atomic_vec2, devices=devices)
    add_function_test(TestAtomic, "test_atomic_vec3", test_atomic_vec3, devices=devices)
    add_function_test(TestAtomic, "test_atomic_vec4", test_atomic_vec4, devices=devices)
    add_function_test(TestAtomic, "test_atomic_mat22", test_atomic_mat22, devices=devices)
    add_function_test(TestAtomic, "test_atomic_mat33", test_atomic_mat33, devices=devices)
    add_function_test(TestAtomic, "test_atomic_mat44", test_atomic_mat44, devices=devices)

    return TestAtomic

if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2)